from ..utils import auto_open
import collections
from time import time
import xml.etree.ElementTree as etree
from pathlib import Path
import logging
logger = logging.getLogger(__name__)

Gene = collections.namedtuple("Gene", "xref species internal_id is_main_isoform")


class SpeciesAnalyser:
    def __init__(self, gene_attr="protId"):
        self.gene_attr = gene_attr
        self.genes = {}
        self.nr_genes_per_species = collections.defaultdict(int)
        self.nr_proteins_per_species = collections.defaultdict(int)

    def add_genome_genes(self, genome_node):
        genome_name = genome_node.get('name', None)
        if genome_name is None:
            genome_name = genome_node.get("NCBITaxId")

        generef_2_xref = {}
        for gene in genome_node.findall('.//{http://orthoXML.org/2011/}gene'):
            gene_id = gene.get('id')
            gene_prot_id = gene.get(self.gene_attr)
            is_main = gene.get('main_isoform', gene_id) == gene_id
            generef_2_xref[gene_id] = Gene(gene_prot_id, genome_name, gene_id, is_main)
            if is_main:
                self.nr_genes_per_species[genome_name] += 1
            self.nr_proteins_per_species[genome_name] += 1
        self.genes.update(generef_2_xref)

    def gene_in_group(self, gene_id):
        self.genes.pop(gene_id)

    def get_singletons(self):
        return self.genes

    def summary(self):
        single = collections.defaultdict(int)
        for g in self.genes.values():
            single[g.species] += 1 if g.is_main_isoform else 0
        return [{'species': g, 
                 'genes': self.nr_genes_per_species[g], 
                 'not_in_group': single[g], 
                 'proteins': self.nr_proteins_per_species[g], 
                 'minor_splice': self.nr_proteins_per_species[g] - self.nr_genes_per_species[g]}
                for g in self.nr_genes_per_species]


def parse_orthoxml(fh, genome_watcher: SpeciesAnalyser):
    taxonomy = {}
    og_level = 0

    def collect_genes(elem):
        genes = 0
        for child in elem.iter():
            if child == elem:
                continue
            if child.tag == "{http://orthoXML.org/2011/}geneRef":
                genes += 1
                if genome_watcher is not None:
                    genome_watcher.gene_in_group(child.get('id'))
            elif child.tag == "{http://orthoXML.org/2011/}orthologGroup":
                genes += child.text
        elem.clear()
        elem.text = genes
        return genes

    logger.info("start mapping of orthoxml formatted input file")
    for event, elem in etree.iterparse(fh, events=('start', 'end')):
        if event == "start":
            if elem.tag == "{http://orthoXML.org/2011/}orthoXML":
                if elem.get('version') != "0.5":
                    raise RuntimeError(f"Expecting orthoXML version 0.5, but is {elem.get('version')}")
            elif elem.tag == '{http://orthoXML.org/2011/}orthologGroup':
                og_level += 1
        elif event == 'end':
            if elem.tag == "{http://orthoXML.org/2011/}orthologGroup":
                og_level -= 1
                data = {'id': elem.get('id'), 'level': taxonomy[elem.get('taxonId')]}
                for child in elem.findall('./{http://orthoXML.org/2011/}score'):
                    data[child.get('id')] = float(child.get('value'))
                data['nr_members'] = collect_genes(elem)
                data['is_roothog'] = og_level == 0
                yield data
                if og_level == 0:
                    elem.clear()
            elif elem.tag == "{http://orthoXML.org/2011/}species":
                if genome_watcher is not None:
                    genome_watcher.add_genome_genes(elem)
                elem.clear()
            elif elem.tag == "{http://orthoXML.org/2011/}taxon":
                taxonomy[elem.get('id')] = elem.get('name')


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--orthoxml", required=True)
    conf = parser.parse_args()
    genome_coverage_stats = SpeciesAnalyser()
    with open(conf.orthoxml, 'rt') as xml:
        for group in parse_orthoxml(xml, genome_coverage_stats):
            print(group)



import os
import logging

from . import _utils_subhog
from . import _infer_subhog
from . import __version__ as fastoma_version
from .logging_setup import setup_logging
logger = logging.getLogger(__name__)

"""

fastoma-infer-subhogs  --input-rhog-folder rhogs_rest/0  --output-pickles "pickle_hogs"  \
    --species-tree  species_tree_checked.nwk -vv --parallel # --msa-write --gene-trees-write
    
"""

def fastoma_infer_subhogs():

    import argparse
    parser = argparse.ArgumentParser(description="checking parameters for FastOMA",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--version", action="version", version="FastOMA v"+fastoma_version)
    parser.add_argument("--input-rhog-folder", required=True, help="Path to the input rootHOG folder.")
    parser.add_argument("--parallel", action='store_true', help="use concurrent parallel per rootHOG")
    parser.add_argument("--species-tree", required=True,
                        help="Path to the input species tree file in newick format")
    parser.add_argument("--output-pickles", required=False, default="pickle_hogs",
                        help="Path to the output folder")

    parser.add_argument("--threshold-dubious-sd", required=False, type=float, default=1/10,
                        help="Threshold to remove proteins in a gene tree due to low species overlap score, not enough evidence for duplication event.") # threshold_dubious_sd
    parser.add_argument("--number-of-samples-per-hog", type=int, default=5,
                        help="Number of representatives (sequences) per HOG. Defaults to ")
    parser.add_argument("--overlap-fragments", required=False, type=float, default=0.15,
                        help="Threshold overlap between two sequences (rows) in MSA to decide whether they are fragments of a gene.")  # overlap_fragments
    parser.add_argument("--gene-rooting-method", required=False, default="midpoint", # gene_rooting_method
                        help="The method used for rooting of gene tree :    midpoint    mad     Nevers_rooting .")
    parser.add_argument("--gene-trees-write", action='store_true',
                        help="writing the all gene trees .")  # the order seems to be nwk_SD_labeled.nwk, dubious_sd0.nwk_SD_labeled.nwk, dubious_sd1.nwk_SD_labeled.nwk
    parser.add_argument("--msa-write", action='store_true',
                        help="writing the raw MSAs (might have more genes that the final gene tree).")
    parser.add_argument("--msa-filter-method",
                        choices=("col-row-threshold", "col-elbow-row-threshold", "trimal"),
                        default="col-row-threshold",
                        help="The method used for filtering MSAs.")
    parser.add_argument("--gap-ratio-row", required=False, type=float, default=0.3,
                        help="For trimming the MSA, the threshold of ratio of gaps for each row.")
    parser.add_argument("--gap-ratio-col", required=False, type=float, default=0.5,
                        help="For trimming the MSA, the threshold of ratio of gaps for each column.")
    parser.add_argument("--min-col-trim", required=False, type=int, default=50,  # todo min rows trim
                        help="min no. columns in msa to consider for filtering")
    parser.add_argument('-v', action="count", default=0, help="Increase verbosity to info/debug")
    conf_infer_subhhogs = parser.parse_args()

    setup_logging(conf_infer_subhhogs.v, conf_infer_subhhogs.parallel)
    logger.debug("Arguments: %s", conf_infer_subhhogs)

    address_rhogs_folder = conf_infer_subhhogs.input_rhog_folder
    # address_rhogs_folder = "./"  # _config.input_rhog_folder
    inferhog_concurrent_on = conf_infer_subhhogs.parallel
    if inferhog_concurrent_on:
        logger.info("parallelization for subhog inference is on.")

    if not os.path.exists(conf_infer_subhhogs.output_pickles):
        os.makedirs(conf_infer_subhhogs.output_pickles)

    pickles_subhog_folder_all = "./" # pickle per taxonomic level

    list_rhog_fastas_files = _utils_subhog.list_rhog_fastas(address_rhogs_folder)
    print("there are ", len(list_rhog_fastas_files), "rhogs in the input folder")

    rhogs_fa_folder = address_rhogs_folder

    list_rhog_fastas_files_rem = _utils_subhog.list_rhog_fastas(address_rhogs_folder)
    print("there are ", len(list_rhog_fastas_files_rem), "rhogs remained in the input folder", list_rhog_fastas_files_rem[:5] )

    hogs_rhog_xml_batch = _infer_subhog.read_infer_xml_rhogs_batch(list_rhog_fastas_files_rem, inferhog_concurrent_on, conf_infer_subhhogs.output_pickles, pickles_subhog_folder_all, rhogs_fa_folder, conf_infer_subhhogs)

    print("finsihed ", address_rhogs_folder)

    threshold_dubious_sd= 0.1


if __name__ == "__main__":
    fastoma_infer_subhogs()


import time
from Bio import AlignIO, SeqIO
from six import StringIO
from ..abstract_cli import AbstractCLI
from .base_aligner import Aligner, AlignmentInput, DataType
from ..options import StringOption, FlagOption, IntegerOption, FloatOption, MultiOption, OptionSet
import tempfile




class ProbConsCLI(AbstractCLI):
    """
    ProbCons low-level command line interface

    :Example:

    ::

        probcons_cli = _ProbConsCLI()
        process = mafft_cli(cmd='mafft args...')
        stdout = mafft_cli.get_stdout()
    """
    @property
    def _default_exe(self):
        return 'probcons'

    # def _set_help(self):
    #     self(help=True, wait=True)
    #     self._help = self.get_stdout()


def set_default_dna_options(aligner):
    """
    Dummy function as sensible default already provided by mafft --auto
    """
    aligner.options = get_default_options()


def set_default_protein_options(aligner):
    """
    Dummy function as sensible default already provided by mafft --auto
    """
    aligner.options = get_default_options()


class ProbCons(Aligner):
    """
    Convenient wrapper for ProbCons multiple sequence aligner

    The wrapper is written as a callable class.
    This can hold data (state) to do with the operation it performs, so it can keep results,
    execution times and other metadata, as well as perform the task.

    This is a basic implementation that can be extended. The important parts are
    __init__ (does the setup) and __call__ (does the work). All
    else are helper methods.

    :Example:

    ::

        callable_wrapper = ProbCons(aln)
        result = callable_wrapper()
        time_taken = callable_wrapper.elapsed_time
        result_again = callable_wrapper.result


    .. note:: There exists an ipython notebook on how to work with wrappers,
         including dealing with non-default parameters.
    """

    def __init__(self, input_, *args, **kwargs):
        super(ProbCons, self).__init__(input_, *args, **kwargs)
        self.options = get_default_options()
        if self.datatype == DataType.DNA:
            set_default_dna_options(self)
        else:
            set_default_protein_options(self)

    def __call__(self, *args, **kwargs):
        """
        Anything to do with calling Mafft should go here.
        If any extra arguments need to be passed they can
        be specified (listed as *args and **kwargs for now).
        """
        start = time.time()  # time the execution
        
        if self.input_type == AlignmentInput.OBJECT:  # different operation depending on what it is
            with tempfile.NamedTemporaryFile(mode='wt') as filehandle:
                SeqIO.write(self.input, filehandle, 'fasta')
                filehandle.seek(0)
                output, error = self._call(filehandle.name, *args, **kwargs)
                
        else:
            output, error = self._call(self.input, *args, **kwargs)

        self.result = self._read_result(output) # store result
        self.stdout = output
        self.stderr = error

        end = time.time()
        self.elapsed_time = end - start
        return self.result
        # End call

    # Any other accessory methods 
    def _call(self, filename, *args, **kwargs):
        """
        Call underlying low level _Mafft wrapper. 
        Options are passed via *args and **kwargs
        [This only covers the simplest automatic
         case]
        """
        self.cli('{} {}'.format(self.command(), filename),
                 wait=True)
        return self.cli.get_stdout(), self.cli.get_stderr()

    def command(self):
        return str(self.options)

    def _read_result(self, output):
        """
        Read back the result.
        """
        fileobj = StringIO(output)
        return AlignIO.read(fileobj, 'fasta')

    def _init_cli(self, binary):
        return ProbConsCLI(executable=binary)


def get_default_options():
    return OptionSet([
        # Algorithm

        # use CLUSTALW output format instead of MFA
        FlagOption('-clustalw', False, active=False),

        # use 0 <= REPS <= 5 (default: 2) passes of consistency transformation
        IntegerOption('-c', 0, active=False),

        # use 0 <= REPS <= 1000 (default: 100) passes of iterative-refinement
        IntegerOption('-ir', 100, active=False),

        # use 0 <= REPS <= 20 (default: 0) rounds of pretraining
        IntegerOption('-pre', 0, active=False),

        # generate all-pairs pairwise alignments
        FlagOption('-pairs', False, active=False),

        #use Viterbi algorithm to generate all pairs(automatically enables - pairs)
        FlagOption('-viterbi', False, active=False),

        # write annotation for multiple alignment to FILENAME
        StringOption('-annot', '', active=False),

        # print sequences in alignment order rather than input order (default: off)
        FlagOption('-a', False, active=False)

    ])



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future.builtins import next
from future.builtins import str
from future import standard_library
standard_library.install_hooks()
from past.builtins import basestring

import io
import collections
from .tools import enum, py2_iterable


class LexError(Exception):

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


@py2_iterable
class Streamer(object):

    """ Wraps an io.StringIO and iterates a byte at a time,
    instead of a line at a time """

    def __init__(self, stream):
        """ _peek always looks ahead 1 position """
        if isinstance(stream, str) or isinstance(stream, basestring):
            stream = io.StringIO(u'{}'.format(stream))
        self.stream = stream
        self._peek = self.stream.read(1)

    def __iter__(self):
        return self

    def __next__(self):
        char = self._peek

        self._peek = self.stream.read(1)

        if self.stream.closed:
            raise StopIteration

        if char == '':
            self.stream.close()
            raise StopIteration

        return char
    def peek(self):
        return self._peek

    def isclosed(self):
        return self.stream.closed

Token = collections.namedtuple('Token', 'typ val')


@py2_iterable
class NewickLexer(object):

    """ Breaks newick stream into lexing tokens:
    Works as a state machine, like Rob Pike's Go text template parser """

    tokens = enum("EOF", "TREE", "LEAF", "SUBTREE", "LABEL",
                  "LENGTH", "SUPPORT", "ENDSUB", "ENDTREE")

    def __init__(self, streamer):
        self.streamer = streamer
        self.token = None
        self.token_buffer = bytearray()
        self.state = self.lex_tree

    def buffer(self, char):
        """ Adds a streamed character to the token buffer """
        self.token_buffer.append(ord(char))

    def eat_spaces(self):
        while self.streamer.peek().isspace():
            next(self.streamer)

    def emit(self, item):
        """ Emits the token buffer's contents as a token; clears the buffer """
        self.token = item
        self.empty_buffer()

    def empty_buffer(self):
        """ Clears the token buffer (python 2 has no bytearray.clear()
        method ) """
        self.token_buffer = self.token_buffer[0:0]

    def __iter__(self):
        return self

    def __next__(self):
        """ Each iteration returns a token. While a token isn't ready,
        advance the state machine one state. """
        while not self.token:
            self.state = self.state()
        token, self.token = self.token, None
        return token

    def pos(self):
        """ Returns position in input stream """
        return self.streamer.stream.tell()

    def truncated_string(self, s, length=60, ellipsis='...'):
        """ Returns a string `s` truncated to maximum length `length`.
        If `s` is longer than `length` it is truncated and `ellipsis` is
        appended to the end. The ellipsis is included in the length.
        If `s` is shorter than `length` `s` is returned unchanged. """
        l = length - len(ellipsis)
        return s[:l] + (s[l:] and ellipsis)

    def lex_tree(self):
        for x in self.streamer:
            if x == '(':
                break

        if self.streamer.isclosed():
            self.emit(Token(self.tokens.EOF, -1))
            raise StopIteration

        self.emit(Token(self.tokens.TREE, ''))
        return self.lex_subtree_start

    def lex_subtree_start(self):
        self.eat_spaces()
        char = self.streamer.peek()

        if char == '(':
            self.emit(Token(self.tokens.SUBTREE, next(self.streamer)))
            return self.lex_subtree_start

        else:
            self.emit(Token(self.tokens.LEAF, None))
            return self.lex_label

    def lex_label(self):
        self.eat_spaces()
        char = self.streamer.peek()
        if char in ('"', "'"):
            next(self.streamer)  # throw away opening quote
            self._match_delimited(char)
        else:
            despacer = {' ': '_'}
            self._match_run(str.isalnum, accepted_chars='-_|.',
                            denied_chars=':,;', replacements=despacer)
        label = self.token_buffer.decode()
        if label == '':
            label = None
        self.emit(Token(self.tokens.LABEL, label))

        return self.lex_length

    def lex_length(self):
        char = self.streamer.peek()
        if char == ':':
            self.streamer.next()  # throw away colon
            self._match_number()
            if len(self.token_buffer) == 0:
                num = None
            else:
                num = float(self.token_buffer)
        else:
            num = None
        self.emit(Token(self.tokens.LENGTH, num))
        return self.lex_subtree_end

    def lex_subtree_end(self):
        self.eat_spaces()
        char = self.streamer.peek()

        if char == ';':
            next(self.streamer)
            self.emit(Token(self.tokens.ENDTREE, ';'))
            return self.lex_tree

        elif char == ',':
            next(self.streamer)
            return self.lex_subtree_start

        elif char == ')':
            next(self.streamer)
            self.emit(Token(self.tokens.ENDSUB, ')'))
            peek = self.streamer.peek()  # is a label or a support value next?
            if peek.isdigit() or peek == '.':
                return self.lex_support
            return self.lex_label

        else:
            raise LexError('Don\'t know how to lex this: {0} ({1})'.format(
                char, self.streamer.stream.tell()))

    def lex_support(self):
        self._match_number()
        if len(self.token_buffer) == 0:
            num = 0.0
        else:
            num = float(self.token_buffer)
        self.emit(Token(self.tokens.SUPPORT, num))
        return self.lex_length

    def _match_delimited(self, delimiter):
        pos = self.pos() - 2  # stream is 2 chars ahead of the opening delimiter
        for char in self.streamer:
            if char == delimiter:
                return
            self.buffer(char)

        buf = self.truncated_string(self.token_buffer.decode())
        msg = ''.join((
            'Unterminated {0}-delimited string starting at '.format(delimiter),
            'position {0}:\n{1}{2}'.format(pos, delimiter, buf)
            ))
        raise LexError(msg)

    def _match(self, predicate, accepted_chars='', denied_chars='',
               replacements=None):
        """ Checks next character in stream. If predicate returns True, or char
        is in `accepted_chars`, advances the stream and returns 1. Else, or if
        the char is in `denied_chars`, doesn't advance the stream and returns 0.
        Replacements is an optional dictionary that can be used to replace the
        streamed character with an alternative (e.g. replace spaces with
        underscores). """

        replacements = (replacements or {})
        char = self.streamer.peek()
        char = replacements.get(char, char)

        if predicate(char) or char in accepted_chars:
            if len(char) == 1:
                self.buffer(char)
            next(self.streamer)  # advance stream
            return 1
        elif char in denied_chars:
            return 0
        else:
            return 0

    def _match_run(self, predicate, **kwargs):
        """ kwargs are `accepted_chars`, `denied_chars` and `replacements` """
        nchars = 0
        try:
            while True:
                matched = self._match(predicate, **kwargs)
                nchars += matched
                if matched == 0:
                    return nchars
        except StopIteration:
            raise LexError('Unexpected end of stream')

    def _match_number(self):
        digits = 0
        self._match(lambda x: False, '-+')
        digits += self._match_run(str.isdigit)
        if self._match(lambda x: False, '.'):
            digits += self._match_run(str.isdigit)

        if digits > 0:
            if self._match(lambda x: False, 'eE'):
                self._match(lambda x: False, '-+')
                self._match_run(str.isdigit)

        else:
            self.empty_buffer()



import collections
import gzip
import os
import pickle
import string
import xml.etree.ElementTree as ET
from pathlib import Path
from datetime import datetime
from Bio import SeqIO
from ete3 import Tree

from ._utils_subhog import read_species_tree
from .transformer import header_transformer
from .zoo.hog import extract_flat_groups_at_level, extract_marker_groups_at_level

from . import __version__ as fastoma_version, logger
from .logging_setup import setup_logging

"""

fastoma-collect-subhogs --pickle-folder pickle_folders  --roothogs-folder omamer_rhogs \
 --gene-id-pickle-file gene_id_dic_xml.pickle --out FastOMA_HOGs.orthoxml  --marker-groups-fasta OrthologousGroups.tsv \
 --roothog-tsv RootHOGs.tsv  --species-tree  species_tree_checked.nwk  -vv
 
"""

# This code collect subhogs and writes outputs.

def iter_hogs(pickle_folder: Path):
    cnt = 0
    nr_hogs = 0
    logger.info("reading pickle files from %s", pickle_folder)
    for root, dirs, files in os.walk(pickle_folder, followlinks=True):
        for f in files:
            if f.startswith('file_') and f.endswith('.pickle'):
                with open(os.path.join(root, f), 'rb') as handle:
                    try:
                        cur = pickle.load(handle)
                    except pickle.PickleError as e:
                        logger.error("Error reading pickle file %s: %s", os.path.join(root, f), e)
                        raise
                nr_hogs += len(cur)
                yield from cur
                cnt += 1
                if cnt % 500 == 0:
                    logger.info("read %d batch pickle files so far, resulting in %d roothogs so far", cnt, nr_hogs)
    logger.info("number of pickle files is %d.", cnt)
    logger.debug("number of hogs in all batches is %d", nr_hogs)


def convert_speciestree_to_orthoxml_taxonomy(tree:Tree):
    cur_id = 1
    name2id = {}
    root = ET.Element("taxonomy")
    for node in tree.traverse(strategy="preorder"):
        try:
            parent_xml = node.up.in_xml
        except AttributeError:
            parent_xml = root
        xml_node = ET.SubElement(parent_xml, "taxon", {'id': str(cur_id), 'name': node.name})
        name2id[node.name] = cur_id
        node.add_feature('in_xml', xml_node)
        cur_id += 1
    return root, name2id


def update_hogids(fam, hog, name2taxid):

    dupCnt = collections.defaultdict(int)
    def _getNextSubId(idx):
        """helper method to return the next number at a given depth of
        duplication (idx)"""
        dupCnt[idx - 1] += 1
        return dupCnt[idx - 1]

    def _encodeParalogClusterId(prefix, nr):
        letters = []
        while nr // 26 > 0:
            letters.append(string.ascii_lowercase[nr % 26])
            nr = nr // 26 - 1
        letters.append(string.ascii_lowercase[nr % 26])
        return prefix + ''.join(letters[::-1])

    def _annotateGroupR(node: ET.ElementTree, og: str, idx: int = 0):
        """create the og attributes at the orthologGroup elements
        according to the naming schema of LOFT. ParalogGroup elements
        do not get own attributes (not possible in the xml schema),
        but propagate their sub-names for the subsequent orthologGroup
        elements."""
        if node.tag == "orthologGroup":
            taxrange = node.find('./property[@name="TaxRange"]')
            taxid = name2taxid[taxrange.get('value')]
            node.set('id', f"{og}_{taxid}")
            node.set('taxonId', str(taxid))
            for child in node:
                _annotateGroupR(child, og, idx)
        elif node.tag == "paralogGroup":
            idx += 1
            nextOG = f"{og}.{_getNextSubId(idx)}"
            for i, child in enumerate(list(node)):
                _annotateGroupR(child, _encodeParalogClusterId(nextOG, i), idx)

    omamer_roothog_id = ":".join(hog.get('id').split("_")[0:2])
    fam_elem = ET.Element("property", {"name": "OMAmerRootHOG", "value": omamer_roothog_id})
    hog.insert(1, fam_elem)
    _annotateGroupR(hog, "HOG:{:07d}".format(fam))
    return hog


def fastoma_collect_subhogs():
    import argparse
    parser = argparse.ArgumentParser(description="collecting all computed HOGs and combine into a single orthoxml")
    parser.add_argument("--version", action="version", version="FastOMA v"+fastoma_version)
    parser.add_argument('--pickle-folder', required=True,
                        help="folder containing the pickle files. Will be searched recursively")
    parser.add_argument('--roothogs-folder', required=True, help="folder containing the omamer roothogs")
    parser.add_argument('--gene-id-pickle-file', required=True,
                        help="file containing the gene-id dictionary in pickle format")
    parser.add_argument('--out', required=False, default="output_hog.orthoxml", help="output filename in orthoxml")
    parser.add_argument('-v', action="count", default=0)
    parser.add_argument('--roothog-tsv', default=None, required=False,
                        help="If specified, a tsv file with the given path will be produced containing the roothog "
                             "assignments as TSV file. In addition, a folder named RootHOGsFasta will be generated"
                             "with one fasta file per inferred RootHOG.")
    parser.add_argument('--marker-groups-fasta', default=None, required=False,
                        help="If specified, a folder named OrthologousFasta and a TSV file with the name provided in "
                             "this argument will be generated that contains single copy groups, i.e. groups which have " 
                             "at most one gene per species. Useful as phylogenetic marker genes to reconstruct species "
                             "trees.")
    parser.add_argument('--species-tree', required=True,
                        help="Path to the species tree used to infer the hogs")
    parser.add_argument('--id-transform', choices=("UniProt", "noop"), default="noop",
                        help="""ID transformer from fasta files to orthoxml / OrthologGroup
                             protein IDs. By default, no transformation will be done. 
                             Existing values are:
                               noop:      No transformation - entire ID of fasta header
                               UniProt:   '>sp|P68250|1433B_BOVIN' --> P68250""")
    conf_collect_subhogs = parser.parse_args()  # conf_collect_subhogs
    setup_logging(conf_collect_subhogs.v)
    logger.debug(conf_collect_subhogs)
    id_transformer = header_transformer(conf_collect_subhogs.id_transform)

    write_hog_orthoxml(conf_collect_subhogs.pickle_folder, conf_collect_subhogs.out, conf_collect_subhogs.gene_id_pickle_file,
                       id_transformer=id_transformer, species_tree=conf_collect_subhogs.species_tree)
    if conf_collect_subhogs.roothog_tsv is not None:
        write_roothogs(Path(conf_collect_subhogs.out), Path(conf_collect_subhogs.roothogs_folder),
                       output_file_roothog_tsv=conf_collect_subhogs.roothog_tsv,
                       id_transformer=id_transformer)
    if conf_collect_subhogs.marker_groups_fasta is not None:
        write_group_files(Path(conf_collect_subhogs.out), Path(conf_collect_subhogs.roothogs_folder),
                          output_file_og_tsv=conf_collect_subhogs.marker_groups_fasta,
                          id_transformer=id_transformer)


def write_hog_orthoxml(pickle_folder, output_xml_name, gene_id_pickle_file, id_transformer, species_tree):
    # todo as input argument/option in nextflow

    # in benchamrk dataset the output prot names should be short
    # tr|A0A0N7KCI6|A0A0N7KCI6_ORYSJ
    # for qfo benchmark, the middle should be written in the file

    orthoxml_file = ET.Element("orthoXML", attrib={"xmlns": "http://orthoXML.org/2011/",
                                                   "origin": "FastOMA "+fastoma_version,
                                                   "originVersion": datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                                                   "version": "0.5"})  #

    with open(gene_id_pickle_file, 'rb') as handle:
        gene_id_name = pickle.load(handle)
        # gene_id_name[query_species_name] = (gene_idx_integer, query_prot_name)
    logger.debug("We read the gene_id_name dictionary with %d items", len(gene_id_name))

    speciestree = read_species_tree(species_tree)
    taxonomy, name2taxid = convert_speciestree_to_orthoxml_taxonomy(speciestree)
    logger.debug("Now creating the header of orthoxml")

   #  #### create the header of orthoxml ####
    for query_species_name, list_prots in gene_id_name.items():
        species_xml = ET.SubElement(orthoxml_file, "species", attrib={"name": query_species_name, "taxonId": str(name2taxid[query_species_name]), "NCBITaxId": "0"})
        database_xml = ET.SubElement(species_xml, "database", attrib={"name": "database", "version": "2023"})
        genes_xml = ET.SubElement(database_xml, "genes")
        for gene in list_prots:
            attribs = {"id": str(gene.numeric_id), "protId": id_transformer.transform(gene.prot_id)}
            if gene.main_isoform is not None:
                attribs["main_isoform"] = str(gene.main_isoform)
            gene_xml = ET.SubElement(genes_xml, "gene", attrib=attribs)
    logger.debug("gene_xml is created.")
    orthoxml_file.append(taxonomy)

    scores = ET.SubElement(orthoxml_file, "scores")
    ET.SubElement(scores, "scoreDef", {"id": "CompletenessScore",
                                       "desc": "Fraction of expected species with genes in the (Sub)HOG"})

    #  #### create the groups of orthoxml   ####
    groups_xml = ET.SubElement(orthoxml_file, "groups")
    for fam, hogs_a_rhog_xml in enumerate(iter_hogs(Path(pickle_folder)), start=1):
        groups_xml.append(update_hogids(fam, hogs_a_rhog_xml, name2taxid))
    logger.debug("converting the xml object to string.")
    with open(output_xml_name, 'wb') as fh:
        ET.indent(orthoxml_file, space='  ', level=0)
        orthoxml = ET.ElementTree(orthoxml_file)
        orthoxml.write(fh, encoding="utf-8", xml_declaration=True, )
    logger.info("orthoxml is written in %s", output_xml_name)


def callback_group_and_omamer(node):
    omamer_node = node.find('./{http://orthoXML.org/2011/}property[@name="OMAmerRootHOG"]')
    omamer = omamer_node.get('value') if omamer_node is not None else 'n/a'
    return {"group_id": node.get('id', 'n/a').split('_')[0],
            "omamer_roothog": omamer}


def write_group_files(orthoxml: Path, roothog_folder: Path, output_file_og_tsv=None, output_fasta_groups=None, id_transformer=None):
    if output_file_og_tsv is None:
        output_file_og_tsv = "OrthologousGroups.tsv"
    if output_fasta_groups is None:
        output_fasta_groups = "OrthologousGroupsFasta"
    output_fasta_groups = Path(output_fasta_groups)
    output_fasta_groups.mkdir(parents=True, exist_ok=True)

    logger.info("Start writing OG tsv and fasta files")
    fasta_format = "fa"  # of the rhogs
    nr_prot_in_groups, nr_groups = 0, 0
    with open(output_file_og_tsv, 'w') as tsv:
        tsv.write("Group\tProtein\n")
        for grp, meta in extract_marker_groups_at_level(orthoxml, protein_attribute="protId", callback=callback_group_and_omamer):
            group_members = {(g.xref, g.species) for g in grp}
            group_name = meta['group_id'].replace("HOG:", "OG_")
            nr_prot_in_groups += len(grp)
            nr_groups += 1
            for gene_xref, gene_species in group_members:
                tsv.write(f"{group_name}\t{gene_xref}\n")

            _write_group_fasta(fasta_format, group_members, group_name, id_transformer, meta, output_fasta_groups,
                               roothog_folder)
    logger.info("writing of %s done. created %d groups containing %d proteins in total",
                    output_file_og_tsv, nr_groups, nr_prot_in_groups)


def write_roothogs(orthoxml: Path, roothog_folder: Path, output_file_roothog_tsv=None, output_fasta_groups=None, id_transformer=None):
    if output_file_roothog_tsv is None:
        output_file_roothog_tsv = "RootHOGs.tsv"
    if output_fasta_groups is None:
        output_fasta_groups = "RootHOGsFasta"
    output_fasta_groups = Path(output_fasta_groups)
    output_fasta_groups.mkdir(parents=True, exist_ok=True)

    logger.info("Start writing RootHOG tsv and fasta files")
    fasta_format = "fa"  # of the rhogs
    nr_prot_in_groups, nr_groups = 0, 0
    with open(output_file_roothog_tsv, 'wt') as tsv:
        tsv.write("RootHOG\tProtein\tOMAmerRootHOG\n")
        for grp, meta in extract_flat_groups_at_level(orthoxml, callback=callback_group_and_omamer):
            group_members = {(g.xref, g.species) for g in grp}
            group_name = meta['group_id']
            # this is the id of the merged roothogs from the placement step
            omamer_roothog = meta['omamer_roothog']
            nr_prot_in_groups += len(grp)
            nr_groups += 1
            for gene_xref, _ in group_members:
                tsv.write(f"{group_name}\t{gene_xref}\t{omamer_roothog}\n")

            _write_group_fasta(fasta_format, group_members, group_name.replace(":", ""), id_transformer, meta, output_fasta_groups,
                               roothog_folder)

    logger.info("writing of %s done. created %d groups containing %d proteins in total",
                    output_file_roothog_tsv, nr_groups, nr_prot_in_groups)


def _write_group_fasta(fasta_format, group_members, group_name, id_transformer, meta, output_fasta_groups,
                       roothog_folder):
    group_seqs = []
    rhog_fasta = roothog_folder / (meta['omamer_roothog'].replace(':', '_') + "." + fasta_format)
    for rec in SeqIO.parse(rhog_fasta, "fasta"):
        orig_id, sp, *rest = rec.id.split("||")
        protid = id_transformer.transform(orig_id)
        if (protid, sp) in group_members:
            rec.id = protid
            rec.description += " [" + sp + "]"
            group_seqs.append(rec)
    with gzip.open(output_fasta_groups / (group_name + "." + fasta_format + ".gz"), 'wt') as fa_out:
        SeqIO.write(group_seqs, fa_out, "fasta")


if __name__ == "__main__":
    fastoma_collect_subhogs()


import os
import time
import logging
import random
from pyparsing import ParseException
import shutil
from Bio import SeqIO


from .parsers import IqtreeParser
from .base_treebuilder import TreeBuilder, AlignmentInput, DataType

from ..abstract_cli import AbstractCLI
from ..options import StringOption, FlagOption, IntegerOption, FloatOption, MultiOption, OptionSet
from ...file_utils import TempFile, TempDir

logger = logging.getLogger(__name__)


class IqtreeCLI(AbstractCLI):
    @property
    def _default_exe(self):
        return ['iqtree-omp', 'iqtree']


def set_default_dna_options(treebuilder):
    """
    Dummy function as sensible default
    """
    treebuilder.options = get_default_options()
    treebuilder.options['-st'].set_value('DNA')


def set_default_protein_options(treebuilder):
    """
    Dummy function as sensible default
    """
    treebuilder.options = get_default_options()
    treebuilder.options['-st'].set_value('AA')


class Iqtree(TreeBuilder):

    def __init__(self, input_, *args, **kwargs):
        super(Iqtree, self).__init__(alignment=input_, *args, **kwargs)
        self.options = get_default_options()
        if self.datatype == DataType.DNA:
            set_default_dna_options(self)
        elif self.datatype == DataType.PROTEIN:
            set_default_protein_options(self)

    def __call__(self, *args, **kwargs):
        """
        Sets up temporary output file location and calls iqtree using _call() function.
        Writes temporary input file if we're working with SeqIO object
        Saves the stdout and stderr and returns
        """
        start = time.time()  # time the execution

        #Need to create temp directory to put raxml output here
        with TempDir() as tmpd:
            if self.input_type is AlignmentInput.OBJECT:  # different operation depending on what it is
                with TempFile() as filename:
                    SeqIO.write(self.input, filename, 'phylip-relaxed') # default interleaved
                    output, error = self._call(filename,tmpd, *args, **kwargs)
            elif self.input_type is AlignmentInput.FILENAME:
                filename = self.input
                output, error = self._call(filename, tmpd, *args, **kwargs)
            else:
                output, error = self._call(None,tmpd, *args, **kwargs)
            self.result = self._read_result(tmpd)  # store result
        self.stdout = output
        self.stderr = error

        end = time.time()
        self.elapsed_time = end - start
        return self.result
        # End call

    # Any other accessory methods
    def _call(self, filename, tmpd, *args, **kwargs):
        """
        Call underlying low level _iqtree wrapper.
        Options are passed via *args and **kwargs
        [This only covers the simplest automatic
         case]
        """
        self.cli('{} -pre {tmp_path} -s {seqfile}'.format(self.command(),
                                                          tmp_path=os.path.join(tmpd, 'tmp_output'),
                                                          seqfile=filename),
                 wait=True)
        return self.cli.get_stdout(), self.cli.get_stderr()

    def command(self):
        return str(self.options)

    def _read_result(self, tmpd):
        """
        Read back the result.
        """

        expected_outfiles = [os.path.join(tmpd, 'tmp_output.iqtree'),
                             os.path.join(tmpd, 'tmp_output.treefile')]
        parser = IqtreeParser()
        try:
            result = parser.to_dict(*expected_outfiles)
        except IOError as ioerr:
            logger.error('Error reading results')
            result = None
        except ParseException as parseerr:
            logger.error('Other parse error', parseerr)
            result = None
        return result

    def _init_cli(self, binary):
        return IqtreeCLI(executable=binary)


def get_default_options():
    return OptionSet([
        # Number of threads
        IntegerOption('-nt', 2, active=True),

        # Set the model for either DNA or AA alignment
        StringOption('-m', '', active=False),

        # Ultrafast bootstrap (>=1000)
        IntegerOption('-bb', 0, active=False),

        # SH-like approximate likelihood ratio test (SH-aLRT)
        IntegerOption('-alrt', 0, active=False),

        # Bootstrap + ML tree + consensus tree (>=100)
        IntegerOption('-b', 0, active=False)
    ])



import os.path
import subprocess
from shutil import which

from . import _utils_roothog, logger
from . import __version__ as fastoma_version
from .logging_setup import setup_logging



"""

fastoma-infer-roothogs --proteomes proteome --hogmap hogmap --out-rhog-folder omamer_rhogs -vv

"""


def fastoma_infer_roothogs():
    import argparse
    parser = argparse.ArgumentParser(description="checking parameters for FastOMA")
    parser.add_argument("--version", action="version", version="FastOMA v"+fastoma_version)
    parser.add_argument("--proteomes", required=True, help="Path to the folder containing the input proteomes")
    parser.add_argument("--splice", help="Path to the folder containing the splice information files")
    parser.add_argument("--hogmap", help="Path to the folder containing the hogmap files")
    parser.add_argument("--out-rhog-folder", required=True, help="Folder where the roothog fasta files are written") #out_rhog_folder
    parser.add_argument('-v', action="count", default=0, help="Increase verbosity to info/debug")
    parser.add_argument('--min-sequence-length', required=False, default=50, type=int,
                        help="minimum sequence length. Shorter sequences will be ignored. (Default=50)")

    parser.add_argument("--mergHOG-ratioMax-thresh", required=False, type=float, default=0.8, help="For merging rootHOGs, threshold of ratioMax ") # mergHOG_ratioMax_thresh
    parser.add_argument("--mergHOG-ratioMin-thresh", required=False, type=float, default=0.9, help="For merging rootHOGs, threshold of ratioMin ") # mergHOG_ratioMin_thresh
    parser.add_argument("--mergHOG-shared-thresh", required=False, type=float, default=10, help="For merging rootHOGs, threshold of number shared proteins ") # mergHOG_shared_thresh
    parser.add_argument("--mergHOG-fscore-thresh", required=False, type=float, default=70, help="For merging rootHOGs, threshold of famlut score shared proteins ") # mergHOG_fscore_thresh
    parser.add_argument("--big-rhog-size", required=False, type=int, default=50*1000, help= "For big rootHOGs, we have different heuristics") # big_rhog_size
    parser.add_argument("--big-fscore-thresh", required=False, type=int, default=95, help="For huge rootHOGs, we have different heuristics, like filtering low family score protiens") # big_fscore_thresh

    conf = parser.parse_args()
    setup_logging(conf.v)
    logger.debug("Arguments: %s", conf)

    # Step 1: Parse input data
    logger.info("Parsing proteomes...")
    species_names, prot_recs_lists, fasta_format_keep = _utils_roothog.parse_proteomes(
        conf.proteomes, conf.min_sequence_length
    )
    logger.info("Processing protein records...")
    prot_recs_all = _utils_roothog.add_species_name_prot_id(prot_recs_lists)

    logger.info("Parsing HOGMAP files...")
    hogmaps, unmapped = _utils_roothog.parse_hogmap_omamer(
        prot_recs_lists, fasta_format_keep, folder=conf.hogmap
    )

    # Step 2: Handle splice isoforms
    isoform_data = None
    if conf.splice is not None and os.path.exists(conf.splice): 
        logger.info("Processing splice variants...")
        isoform_data = _utils_roothog.handle_splice_variants(species_names, hogmaps, conf.splice)
        _utils_roothog.write_selected_isoforms(isoform_data, prot_recs_lists)
        hogmaps = isoform_data['filtered_hogmaps']

    # Step 3: Group proteins into root HOGs
    logger.info("Grouping proteins into root HOGs...")
    # rhogs_prots = _utils_roothog.group_prots_roothogs(hogmaps)
    # rhogs_prots = _utils_roothog.handle_singleton(rhogs_prots, hogmaps, conf)
    # rhogs_prots = _utils_roothog.merge_rhogs2(hogmaps, rhogs_prots, conf)
    # rhogs_prots = _utils_roothog.filter_big_roothogs(hogmaps, rhogs_prots, conf)

    rhogs_prots = _utils_roothog.create_root_hogs(hogmaps, conf)


    # Step 4: Save results
    logger.info("Saving results...")
    _utils_roothog.save_gene_id_mapping(prot_recs_all, isoform_data)
    # _utils_roothog.write_root_hogs(rhogs_prots, prot_recs_all, conf.out_rhog_folder)
    
    logger.info(f"Successfully created {len(rhogs_prots)} root HOGs")


    min_rhog_size = 2
    rhogid_written_list = _utils_roothog.write_rhog(rhogs_prots, prot_recs_all, conf.out_rhog_folder, min_rhog_size)
    linclust_available = which("mmseqs")  # True #
    # if memseqs is not installed the output will be empty / None
    if linclust_available:
        try:
            res = subprocess.run([linclust_available, "-h"], check=True, text=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            logger.error("mmseqs linclust is not working: %s", e)
            logger.error("Skipping clustering of unmapped singletons")
            linclust_available = False
    if linclust_available:
        singleton_unmapped_path = "singleton_unmapped.fa"
        cnt_unmapped_singleton = _utils_roothog.collect_unmapped_singleton(rhogs_prots, unmapped, prot_recs_all, unmapped_singleton_fasta=singleton_unmapped_path)
        if cnt_unmapped_singleton:
            cluster_file = _utils_roothog.run_linclust(fasta_to_cluster="singleton_unmapped.fa")
            num_clusters = _utils_roothog.write_clusters(cluster_file, conf.out_rhog_folder, min_rhog_size)
            logger.debug("we wrote %d new clusters with linclust", num_clusters)
    else:
        logger.info("mmseqs linclust / easy-cluster not available, skipping clustering of unmapped and singletons")


if __name__ == "__main__":
    fastoma_infer_roothogs()



# Author: Ivana Pilizota
# Date: 1 November 2016

import logging
import os
import time

from Bio import SeqIO
from pyparsing import ParseException
import tempfile

from .base_treebuilder import TreeBuilder, AlignmentInput, DataType, WrapperError
from .parsers import FasttreeParser

from ..abstract_cli import AbstractCLI
from ..options import OptionSet, StringOption, IntegerOption
from ...file_utils import TempFile, TempDir

logger = logging.getLogger(__name__)




class FasttreeCLI(AbstractCLI):
    @property
    def _default_exe(self):
        return ['fasttree', 'FastTree']


def set_default_dna_options(treebuilder):
    """
    Dummy function as sensible default
    """
    treebuilder.options = get_default_options()


def set_default_protein_options(treebuilder):
    """
    Dummy function as sensible default
    """
    treebuilder.options = get_default_options()


class Fasttree(TreeBuilder):

    def __init__(self, alignment, *args, **kwargs):
        self.options = get_default_options()
        super(Fasttree, self).__init__(alignment=alignment, *args, **kwargs)
        if self.input is not None:
            if self.datatype == DataType.DNA:
                set_default_dna_options(self)
            else:
                set_default_protein_options(self)

    def __call__(self, *args, **kwargs):
        """
        Sets up temporary output file location and calls FastTree using _call() function.
        Writes temporary input file if we're working with SeqIO object
        Saves the stdout and stderr and returns
        """
        start = time.time()  # time the execution
        if self.input_type == AlignmentInput.OBJECT:  # different operation depending on what it is
            with tempfile.NamedTemporaryFile(mode='wt') as fh:
                SeqIO.write(self.input, fh, 'fasta') # default interleaved # 'phylip-relaxed'
                fh.seek(0)
                output, error = self._call(fh.name, *args, **kwargs)
                self.result = self._read_result(output, error)  # store result
        else:
            filename = os.path.abspath(self.input)
            output, error = self._call(filename, *args, **kwargs)
            self.result = self._read_result(output, error)  # store result

        end = time.time()
        self.elapsed_time = end - start
        return self.result["tree"]
        # End call

    # Any other accessory methods
    def _call(self, filename, *args, **kwargs):
        """
        Call underlying low level FastTree wrapper.
        Options are passed via *args and **kwargs
        [This only covers the simplest automatic
         case]
        """
        #hard code tmp_output as the output name since we don't save it anyway
        #self.cli('{} -log {log_output} {seqfile} > {tmp_path}'.format(self.command(), tmp_path=os.path.join(tmpd,'tmp_output'), log_output=logfile, seqfile=filename), wait=True)
        self.cli('{} {seq_file}'.format(self.command(), seq_file=filename), wait=True)
        self.returncode = self.cli.process.returncode

        if self.returncode != 0:
            self.stderr = self.cli.get_stderr()
            last_error_line = self.stderr.split('\n')[-1].strip()
            msg = f"Fasttree failed on {filename}: {last_error_line}"
            logger.error(msg)
            raise WrapperError(msg, self.stderr)

        return (self.cli.get_stdout(), self.cli.get_stderr())

    def command(self):
        return str(self.options)

    def _read_result(self, stdout, stderr):
        """
        Read back the result.
        """
        parser = FasttreeParser()

        try:
            parser.parse(tree=stdout, other=stderr)
            result = parser.to_dict()
        except IOError as ioerr:
            logger.error('Error reading results')
            result = None
        except ParseException as parseerr:
            logger.error('Other parse error', parseerr)
            result = None

        return result

    def _init_cli(self, binary):
        return FasttreeCLI(executable=binary)


def get_default_options():

    return OptionSet([
        # Algorithm

        # Set datatype to DNA (nt) or AA alignment: AA by default. If set to True will assume DNA format.
        StringOption('-nt', active=False),

        # Set the WAG model for AA alignment. Default Jones-Taylor-Thorton
        StringOption('-wag', active=False),

        # Set the GTR model for nt alignment. Default Jones-Taylor-Thorton
        StringOption('-gtr', active=False),

        # Set the gamma model. Default Jones-Taylor-Thorton
        StringOption('-gamma', active=False),

        # Specify the number of rate categories of sites. Default 20.
        IntegerOption('-cat', 20, active=False),

        IntegerOption('-seed',1234, active=False),

        # Specify starting tree
        StringOption('-intree', '', active=False),

        # Speed up the neighbor joining phase & reduce memory usage (recommended for >50,000 sequences)
        StringOption('-fastest', active=False),
        # allow spaces and other restricted characters (but not ' ) in sequence names and quote names in the output tree (fasta input only; FastTree will not be able to read these trees back in)
        StringOption('-quote', active=True),

        #-quote -- quote sequence names in the output and allow spaces, commas,  parentheses, and colons in them but not ' characters (fasta files only)\n"
        StringOption('-quote', active=False),

        # Set the number of rounds of maximum-likelihood NNIs. Deafault 4*log2(N), N = the number of unique sequences
        IntegerOption('-mlnni', 0, active=False),

    ])




import csv
import itertools
from pathlib import Path
import subprocess
from itertools import chain

import networkx as nx
from Bio import SeqIO
import pickle
from os import listdir
import os
import sys

from .zoo.unionfind import UnionFind
from . import logger
import collections

filter_nonchild_rootHOG= False
mmseqs_executable_path ="mmseqs"

HOGMapData = collections.namedtuple("HOGMapData", ("hogid", "score", "seqlen", "subfamily_medianseqlen"))
Gene = collections.namedtuple("Gene", ("numeric_id", "prot_id", "main_isoform"), defaults=(None,) )
    

"""UnionFind.py

Union-find data structure. Based on Josiah Carlson's code,
http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/215912
with significant additional changes by D. Eppstein and
Adrian Altenhoff.
"""


class UnionFind(object):
    """Union-find data structure.

    Each unionFind instance X maintains a family of disjoint sets of
    hashable objects, supporting the following two methods:

    - X[item] returns a name for the set containing the given item.
      Each set is named by an arbitrarily-chosen one of its members; as
      long as the set remains unchanged it will keep the same name. If
      the item is not yet part of a set in X, a new singleton set is
      created for it.

    - X.union(item1, item2, ...) merges the sets containing each item
      into a single larger set.  If any item is not yet part of a set
      in X, it is added to X as one of the members of the merged set.
    """

    def __init__(self, elements=None):
        """Create a new union-find structure.

        If elements is not None, the structure gets initialized
        with each element as a singleton component.

        :param elements: an iterable to initialize the structure.
        """

        self.weights = {}
        self.parents = {}
        if elements is not None:
            for elem in iter(elements):
                self.parents[elem] = elem
                self.weights[elem] = 1

    def __getitem__(self, obj):
        """return the name of set which contains obj.

        :param obj: the query object

        :SeeAlso: :meth:`find`"""
        return self.find(obj)

    def find(self, obj):
        """Find and return the name of the set containing the obj.

        If the object is not found in any set, a new singleton set
        is created that holds only this object until it is further merged."""

        # check for previously unknown obj. If unknown, add it
        # as a new cluster
        if obj not in self.parents:
            self.parents[obj] = obj
            self.weights[obj] = 1
            return obj

        # find path of objects leading to the root
        path = [obj]
        root = self.parents[obj]
        while root != path[-1]:
            path.append(root)
            root = self.parents[root]

        # compress the path and return
        for ancestor in path:
            self.parents[ancestor] = root
        return root

    def remove(self, obj):
        """Remove an object from the sets.

        Removes an object entirly from the datastructure. The
        containing set will shrink by this one element.

        :Note: If one tries to accessed it afterwards using
            :meth:`find`, it will be created newly and put as a
            singleton.
        """
        if obj not in self.parents:
            return
        comp = self.find(obj)
        self.weights[comp] -= 1
        self.parents.pop(obj)

    def __iter__(self):
        """Iterate through all items ever found or unioned by this structure."""
        return iter(self.parents)

    def union(self, *objects):
        """Find the sets containing the objects and merge them.

        any number of objects can be passed to this method and
        all of them will be merged into one set containing at
        least these objects.

        :param objects: the objects to be merged. they have to be all
            hashable. If they haven't been initialy added to the UnionFind
            datastructre at instantiation time, they are added at this point
            in time.
        """
        roots = [self[x] for x in objects]
        heaviest = max([(self.weights[r], r) for r in roots], key=lambda x: x[0])[1]
        for r in roots:
            if r != heaviest:
                self.weights[heaviest] += self.weights[r]
                self.parents[r] = heaviest

    def get_components(self):
        """return a list of sets corresponding to the connected
        components of the structure."""
        comp_dict = collections.defaultdict(set)
        for elem in iter(self):
            comp_dict[self[elem]].add(elem)
        comp = list(comp_dict.values())
        return comp


def parse_proteomes(folder=None, min_sequence_length=0):  # list_oma_species
    """
    parsing fasta files of proteins located in /proteome/
    using Bio.SeqIO.parse
    Each fasta file is for one species.  The file name is the species name.
    All proteomes should be with the same extension.
    output: prot_recs_lists: a dic with key as species name and  its value list of Biopython record of species.
    """
    if folder is None:
        folder = "./proteome"

    fasta_format_keep = ""
    project_files = listdir(folder)
    logger.debug("using '%s' as proteome folder, found %d files", folder, len(project_files))
    species_names = []  # query/input species name based on the file name
    for file in project_files:
        species_name, ext = file.rsplit('.', 1)
        logger.debug("%s: %s | %s", file, species_name, ext)
        if ext in ("fa", "fasta"):
            species_names.append(species_name)
            fasta_format_keep = ext  # last one is stored either fa or fasta

    species_names.sort()   # sort the species names alphabetically -> reproducible results
    # todo accept all fasta formats in the input prtoeome folder, fasta, fa, fna, ..
    prot_recs_lists = {} # key: species name, value is a dic of query protein Biopython records. # 'MYCGE': [SeqRecord(seq=Seq('MDFDK
    #smallprot_recs_lists ={}
    for species_name in species_names:
        prot_address = os.path.join(folder, species_name + "." + fasta_format_keep)
        prots_record = list(rec for rec in SeqIO.parse(prot_address, "fasta") if len(rec) >= min_sequence_length)
        #prots_record_small = list(rec for rec in SeqIO.parse(prot_address, "fasta") if len(rec) < min_sequence_length)
        # logger.debug(prots_record)
        logger.info(f"{species_name} contains {len(prots_record)} that are at least {min_sequence_length} long.")
        prot_recs_lists[species_name] = prots_record
        #smallprot_recs_lists[species_name]=prots_record_small

    logger.info("There are %d species in the proteome folder.", len(species_names))
    return species_names, prot_recs_lists, fasta_format_keep #, smallprot_recs_lists



def create_root_hogs(hogmaps, conf):
    """Create root HOGs from HOG mappings"""
    
    # Initial grouping
    rhogs_prots = group_prots_by_roothogs(hogmaps)
    logger.info(f"Initial grouping: {len(rhogs_prots)} root HOGs")
    
    # Handle singletons
    rhogs_prots = resolve_singletons(rhogs_prots, hogmaps, conf)

    rhogs_prots = merge_rhogs2(hogmaps, rhogs_prots, conf)
    rhogs_prots = filter_big_roothogs(hogmaps, rhogs_prots, conf)
    
    # # Merge similar HOGs
    # if conf.enable_merging:
    #     rhogs_prots = merge_similar_hogs(rhogs_prots, hogmaps, conf)
    #
    # # Filter large HOGs
    # rhogs_prots = filter_large_hogs(rhogs_prots, hogmaps, conf)
    
    return rhogs_prots


def handle_splice_variants(species_names, hogmaps, splice_folder):
    """Handle splice variants and return filtered data"""
    
    # Parse isoform files
    isoform_by_gene = parse_isoform_files(species_names, splice_folder)
    
    # Select best isoforms
    selected_isoforms, non_selected = select_best_isoforms(
        species_names, isoform_by_gene, hogmaps
    )
    
    # Filter hogmaps to keep only selected isoforms
    filtered_hogmaps = filter_hogmaps_by_isoforms(hogmaps, non_selected)
    
    return {
        'isoform_by_gene': isoform_by_gene,
        'selected_isoforms': selected_isoforms,
        'non_selected': non_selected,
        'filtered_hogmaps': filtered_hogmaps
    }


def add_species_name_prot_id(prot_recs_lists):
    """
    adding the name of species to each protein record
        - based on file name
    adding protein idx number, integer needed by xml format
    output:  prot_recs_all =  {'MYCGE': {'sp|P47500|RF1_MYCGE||MYCGE||1000000001': SeqRecord(seq=
    """
    start_num_prot = int(1e9)
    start_num_prot_per_sp = int(1e6) #
    prot_recs_all = {} # {'MYCGE': {'sp|P47500|RF1_MYCGE||MYCGE||1000000001': SeqRecord(seq=
    species_idx = -1
    for species_name, prot_recs_list in prot_recs_lists.items():
        species_idx += 1
        prot_idx = start_num_prot + species_idx * start_num_prot_per_sp
        prot_recs_all[species_name]={}
        for prot_rec in prot_recs_list:
            prot_idx+=1
            prot_name= prot_rec.id

            if len(prot_name) > 230:
                logger.info("We are truncating the prot name as it may be problematic for mafft, " + str(prot_name))
                prot_name = prot_name[:230]

            prot_name_new = prot_name+ "||"+species_name+"||"+str(prot_idx) # orthoxml file needs an integer as
            prot_rec.id = prot_name_new
            prot_recs_all[species_name][prot_name] = prot_rec
    return prot_recs_all


def parse_hogmap_omamer(proteomes, fasta_format_keep, folder=None):
    """
     function for parsing output of omamer (hogmap files) located in /hogmap/
    Each hogmap file correspond to one fasta file of species, with the same name.
    Note that some records of fasta may removed in hogmap, probably becuase of being so short.
    hogmap file example:
    # qseqid	hogid	hoglevel	family_p	family_count	family_normcount	subfamily_score	subfamily_count	qseqlen	subfamily_medianseqlen	qseq_overlap
    # sp|O66907|ATPA_AQUAE	HOG:C0886513.1b	Eukaryota	696.5519485850672	187	0.37302594463771466	0.2643067275644273	120	504	551	0.8230616302186878
    output is dic of dic for all species:
    """
    if folder is None:
        folder = "./hogmap"

    hogmaps = {}
    unmapped = collections.defaultdict(list)
    for species_name, prot_reqs in proteomes.items():
        proteome = set(rec.id.split('||', 1)[0] for rec in prot_reqs)
        hogmap_address = os.path.join(folder, species_name + "." + fasta_format_keep + ".hogmap")
        cur_species_hogmap = collections.defaultdict(list)
        with open(hogmap_address, 'rt') as hogmap_file:
            reader = csv.DictReader((line for line in hogmap_file if not line.startswith('!')),
                                    dialect="excel-tab")
            for row in reader:
                if row['hogid'] == "N/A":
                    unmapped[species_name].append(row['qseqid'])
                elif row['qseqid'] not in proteome:
                    logger.debug(f"Ignoring {row} [{species_name}] from hogmap as not in proteome")
                else:
                    cur_species_hogmap[row['qseqid']].append(
                        HOGMapData(row['hogid'], row['family_p'], row['qseqlen'], row['subfamily_medianseqlen']))
        hogmaps[species_name] = cur_species_hogmap
        logger.info("hogmap %s: %d proteins mapped to %d hogs, %d not mapped",
                        species_name, len(cur_species_hogmap), sum(len(x) for x in cur_species_hogmap.values()),
                        len(unmapped[species_name]))

    logger.info("There are %d species in the hogmap folder.", len(hogmaps))
    return hogmaps, unmapped

def group_prots_by_roothogs(hogmaps):
    """Group proteins by their root HOG assignments"""
    
    rhogs_prots = collections.defaultdict(list)
    
    for species_name, prots_map in hogmaps.items():
        for prot_id, prot_mappings in prots_map.items():
            # Get best mapping (first one, as they're sorted by score)
            best_mapping = prot_mappings[0]
            rhogid = extract_root_hog_id(best_mapping.hogid)
            rhogs_prots[rhogid].append((species_name, prot_id))
    
    logger.info(f"Grouped proteins into {len(rhogs_prots)} root HOGs")
    return dict(rhogs_prots)

def extract_root_hog_id(hogid):
    """Extract root HOG ID from full HOG ID"""
    return hogid.split(".")[0].split(":")[1]


def resolve_singletons(rhogs_prots, hogmaps, conf):
    """Resolve singleton HOGs by checking alternative mappings"""
    
    singletons = [(rhog, prots[0]) for rhog, prots in rhogs_prots.items() if len(prots) == 1]
    resolved_count, remained_count = 0, 0
    singletons_remained = collections.defaultdict(set)

    for singleton_rhog, (species, prot) in singletons:
        prot_mappings = hogmaps[species][prot]
        
        # Check alternative mappings
        could_resolve = False
        for alt_mapping in prot_mappings[1:]:
            if float(alt_mapping.score) < conf.mergHOG_fscore_thresh:
                continue
                
            alt_rhog = extract_root_hog_id(alt_mapping.hogid)
            
            # If alternative HOG has multiple members, move protein there
            if alt_rhog in rhogs_prots and len(rhogs_prots[alt_rhog]) > 1:
                rhogs_prots[alt_rhog].append((species, prot))
                del rhogs_prots[singleton_rhog]
                resolved_count += 1
                could_resolve = True
                break
        if not could_resolve:
            remained_count += 1
            for alt_mapping in prot_mappings[1:]:
                if float(alt_mapping.score) < conf.mergHOG_fscore_thresh:
                    continue
                singletons_remained[extract_root_hog_id(alt_mapping.hogid)].add((species, prot))    
    
    logger.info(f"Resolved {resolved_count} singleton HOGs based on omammer multi-hits.")
    logger.info(f"{remained_count} singleton HOGs remained before merging based on minor mappings.")
    
    # get lookup from prot -> roothog (to be used for deleting a new cluster assignment)
    prots_rhogs_dic = {sp_prots[0]: rhog for rhog, sp_prots in rhogs_prots.items() if len(sp_prots) == 1}

    # Merge remaining singletons into groups based on shared alternative mappings of multi-hits
    clusters = UnionFind()
    for rhog, prot_set in singletons_remained.items():
        clusters.union(*prot_set)

    nr_new_rhogs, nr_prot_in_new_rhogs = 0, 0
    for cc in clusters.get_components():
        if len(cc) <= 1:
            continue

        nr_new_rhogs += 1
        # check that indeed the proteins are singletons in the rhog_prots dict
        if any(len(rhogs_prots.get(prots_rhogs_dic[prot], [])) > 1 for prot in cc):
            raise RuntimeError("issue 15219325")
            
        for prot in cc:
            rhog = prots_rhogs_dic[prot]
            if rhog in rhogs_prots:
                del rhogs_prots[rhog]
        cc = list(cc)
        rhog = prots_rhogs_dic[cc[0]]
        rhogs_prots[rhog] = cc
        nr_prot_in_new_rhogs += len(cc)
    logger.info(f"We merged {nr_prot_in_new_rhogs} in {nr_new_rhogs} additional groups with >1 members.")

    counter_rhog_singleton = 0
    counter_not_singleton  = 0
    for rhogid, spec_prot_list in rhogs_prots.items():
        if len(set(spec_prot_list)) > 1:
            counter_not_singleton += 1
        else:
            counter_rhog_singleton += 1
    logger.info(f"Now, we have {counter_not_singleton} rootHOGs with >1 proteins and {counter_rhog_singleton} singleton rootHOGs")
    return rhogs_prots

def save_gene_id_mapping(prot_recs_all, isoform_data=None, out_path="gene_id_dic_xml.pickle"):
    """Save comprehensive gene ID mapping including isoform information"""
    
    gene_mapping = {}
    for species_name, prot_dict in prot_recs_all.items():
        species_data = []

        isoform_to_main = {}
        if isoform_data and species_name in isoform_data['selected_isoforms']:
            isoform_to_main = {gene: main for main, genes in zip(
                    isoform_data['selected_isoforms'][species_name], isoform_data['isoform_by_gene'][species_name]
                ) for gene in genes}
        
        id2numeric = {}
        for prot_name, prot_rec in prot_dict.items():
            parts = prot_rec.id.split('||')
            original_id = parts[0]
            numeric_id = int(parts[2])
            id2numeric[original_id] = numeric_id
        
        species_data = [Gene(num, orig_id, id2numeric.get(isoform_to_main.get(orig_id))) for orig_id, num in id2numeric.items()]
        gene_mapping[species_name] = species_data
    
    # Save to pickle file
    with open(out_path, 'wb') as handle:
        pickle.dump(gene_mapping, handle, protocol=pickle.HIGHEST_PROTOCOL)
    
    logger.info(f"Saved comprehensive gene ID mapping to {out_path}")
    return gene_mapping


def filter_big_roothogs(hogmaps, rhogs_prots, conf_infer_roothogs):


    prots_list_big = []  # as backup
    rhogids_big = []
    for rhogid, sp_prot_list in rhogs_prots.items():
        if len(sp_prot_list) > conf_infer_roothogs.big_rhog_size:
            prots_list_big.append(sp_prot_list)  # (species_name,prot_id)
            rhogids_big.append(rhogid)
            logger.info(
                "a big rootHOG was found " + rhogid + " with " + str(len(sp_prot_list)) + " query proteins.")
    logger.info("There are " + str(len(rhogids_big)) + " big rootHOGs.")

    for rhogid in rhogids_big:

        sp_prot_list = rhogs_prots[rhogid]
        sp_prot_list_filt = []
        hogids = []

        # removing proteins with low omamer score from big rootHOG
        for species_name, prot_id in sp_prot_list:
            prot_maps = hogmaps[species_name][prot_id]
            # if len(prot_maps) > 1:  # probably for big rootHOG, there won't be multi-hits
            #     scores = [float(i[1]) for i in prot_maps]  # (hogid,score,seqlen,subfamily_medianseqlen)
            #     hogids = [i[0] for i in prot_maps]
            #     max_score = max(scores)
            #     max_index = scores.index(max_score)
            #     hogid = hogids[max_index]
            # else:
            #     hogid = prot_maps[0][0]
            #     max_score = float(prot_maps[0][1])
            # todo: highest normcount or pavalue , default of omamer ?
            hogid = prot_maps[0][0]
            max_score = float(prot_maps[0][1])

            if max_score > conf_infer_roothogs.big_fscore_thresh:
                sp_prot_list_filt.append((species_name, prot_id))
                hogids.append(hogid)

        logger.info("For big rootHOG " + rhogid + ", " + str(
            len(sp_prot_list_filt)) + " proteins left after filtering with threshold " + str(conf_infer_roothogs.big_fscore_thresh))

        if len(sp_prot_list_filt) < conf_infer_roothogs.big_rhog_size:
            sp_prot_list_filt2 = sp_prot_list_filt

        elif  filter_nonchild_rootHOG :  # removing proteins that are mapped to rootHOG (= HOGC123 not the levelsHOGC123.1a) from big rootHOG
            #hogids2 = []
            sp_prot_list_filt2 = []
            for prot_idx, sp_prot in enumerate(sp_prot_list_filt): #[('UP000192223_224129', 'tr|A0A1W4WAU6|A0A1W4WAU6_AGRPL'), ('UP000192223_224129', 'tr|A0A1W4WU99|A0A1W4WU99_AGRPL'),
                species , protein = sp_prot
                prot_maps =hogmaps[species][protein]
                hogid = prot_maps[0][0]
                if len(hogid.split(".")) > 1: # if mapped to subHOG (not to the rootHOG)
                    sp_prot_list_filt2.append(sp_prot)
                    #hogids2.append(hogid)
        else:
            sp_prot_list_filt2 = sp_prot_list_filt
        logger.info("For big rootHOG " + rhogid + ", " + str(len(sp_prot_list_filt2)) + " proteins left after removing non-child subhogs if activate: "+str(filter_nonchild_rootHOG))

        if len(sp_prot_list_filt2):
            rhogs_prots[rhogid] = sp_prot_list_filt2
        else:
            del rhogs_prots[rhogid]

    return rhogs_prots


def write_rhog(rhogs_prot_records, prot_recs_all, address_rhogs_folder, min_rhog_size=1):
    # max_rhog_size =1e12
    #address_rhogs_folder = folder + "rhog"
    logger.info("Writing Sequences of roothogs are fasta file in " + address_rhogs_folder)
    if not os.path.exists(address_rhogs_folder):
        os.mkdir(address_rhogs_folder)


    rhogid_written_list = []
    for rhogid, rhog_prots in rhogs_prot_records.items():
        rhog_recs = []
        for (species_name, prot_name) in rhog_prots: #
            if prot_name in prot_recs_all[species_name]: # some small prots are removed in the begining min_sequence_length
                prot_rec = prot_recs_all[species_name][prot_name]
                rhog_recs.append(prot_rec)


        if min_rhog_size <= len(rhog_recs):  # <= max_rhog_size:
            # todo add the release id   to file names  rhogids_list[:2] > ['HOG:C0884658', 'HOG:C0709155']
            SeqIO.write(rhog_recs, address_rhogs_folder + "/HOG_" + rhogid + ".fa", "fasta")
            rhogid_written_list.append(rhogid)
        else:
            logger.debug("The roothog " +str(rhogid)+" was too small with size of "+str(len(rhog_recs))+" which is smaller than threshold "+str(min_rhog_size))
            #logger.debug(" issue 12314405 to check" )
            #for prot1 in rhog_recs:
                #logger.debug("we are removing due to omamer signleton hog |*|" + str(prot1.id))

    logger.info("Writing Sequences of " + str(len(rhogid_written_list)) + " roothogs finished.")

    return rhogid_written_list


def find_rhog_candidate_pairs(hogmaps, rhogs_prots, conf_infer_roothogs): # rhogs_prots
    """find pairs of roothogs that could be merged based on the number
    of shared assignments and sizes.
    """
    pair_rhogs_count = collections.defaultdict(int)
    rhogs_size = collections.defaultdict(int)
    for species_name, prt_prot_maps in hogmaps.items():
        for prot, prot_maps in prt_prot_maps.items():
            # prot_maps= [HOGMapData(hogid='HOG:E0315075.2a.2b.13b', score='3115.977589303038', seqlen='574', subfamily_medianseqlen='532'),
            rhogs = []
            for prot_map in prot_maps: #
                # prot_map: HOGMapData record
                if float(prot_map.score) > conf_infer_roothogs.mergHOG_fscore_thresh:
                    rhogid = extract_root_hog_id(prot_map.hogid)
                    rhogs_size[rhogid] += 1
                    rhogs.append((rhogid, float(prot_map.score)))

            rhogs.sort()  # once sorted, iteration will yield sorted tuples in pairs
            for (hogi, score_i), (hogj, score_j) in itertools.combinations(rhogs, 2):
                # skip pairs of hogs if one of them has not been identified as the best
                # scoring hog in omamer. this could lead to big hairball graphs
                if not (hogi in rhogs_prots and hogj in rhogs_prots):
                    continue
                pair = (hogi, hogj)
                pair_rhogs_count[pair] += 1

    print(len(pair_rhogs_count))
    logger.debug("There are " + str(len(pair_rhogs_count)) + " pairs of rhogs.")

    candidates_pair = []
    for (hogi, hogj), count_shared in pair_rhogs_count.items():
        if hogi in rhogs_size and hogj in rhogs_size:  # during previous functions, we might
            ratioMax = count_shared / max(rhogs_size[hogi], rhogs_size[hogj])
            ratioMin = count_shared / min(rhogs_size[hogi], rhogs_size[hogj])
            # todo: big_rhog_size make it as a ratio, could be problometic when we have many more species -> merging too much
            condition_merge = (max(rhogs_size[hogi], rhogs_size[hogj]) < conf_infer_roothogs.big_rhog_size / 6 and \
                                ((ratioMax > conf_infer_roothogs.mergHOG_ratioMax_thresh or ratioMin > conf_infer_roothogs.mergHOG_ratioMin_thresh) and count_shared > conf_infer_roothogs.mergHOG_shared_thresh))  \
                              and rhogs_size[hogi] < conf_infer_roothogs.big_rhog_size / 6 and rhogs_size[hogj] < conf_infer_roothogs.big_rhog_size / 6
            if condition_merge:
                if rhogs_size[hogi] >= rhogs_size[hogj]:
                    candidates_pair.append((hogi, hogj))  # bigger first
                else:
                    candidates_pair.append((hogj, hogi))
                # print(hogi,"(",rhogs_size[hogi],")",hogj,"(",rhogs_size[hogj],")", count_shared,round(ratioMax,2),round(ratioMin,2))
            # D0651051 ( 59 ) D0658569 ( 180 ) 62 0.34 1.05 # todo why bigger than 1?
            # D0646495 ( 2 ) D0631227 ( 14 ) 25 1.79 12.5

    logger.info("There are %d candidate pairs of rhogs for merging.", len(candidates_pair))
    return candidates_pair


def cluster_rhogs(candidates_pair):
    cluster_rhogs_obj = UnionFind()
    for pair in candidates_pair:
        cluster_rhogs_obj.union(pair[0], pair[1])
    cluster_rhogs_sets = cluster_rhogs_obj.get_components()

    cluster_rhogs_list = [ ]
    for cl in cluster_rhogs_sets:
        cluster_rhogs_list.append(list(cl))

    return cluster_rhogs_list
#
# def cluster_rhogs_old(candidates_pair): # doesnt work correctly
#     # init
#     all_hog_raw = []
#     for pair in candidates_pair:
#         all_hog_raw.append(pair[0])
#         all_hog_raw.append(pair[1])
#     all_hog = list(set(all_hog_raw))
#
#     allcc = []  # connected compoenets
#     for hog in all_hog:
#         allcc.append([hog])
#
#     dic_where = {}
#     for idx, cc in enumerate(allcc):
#         dic_where[cc[0]] = idx  # in the beginning, each inner list has only on element, a unique hog
#
#     # print(len(all_hog_raw),len(all_hog),len(allcc),allcc[:2])
#     logger.debug("There are " + str(len(all_hog)) + " all_hog.")
#
#     # print(dic_where)
#     for pair in candidates_pair:
#         # print(pair)
#         idx_0 = dic_where[pair[0]]
#         idx_1 = dic_where[pair[1]] # this will be problametic for sequence of pairs
#
#         dic_where[pair[1]] = idx_0
#
#         # print(idx_0)
#         allcc[idx_0] += allcc[idx_1]
#         allcc[idx_1] = []
#
#         # print(dic_where,allcc)
#
#     cluster_rhogs_list = [i for i in allcc if len(i) > 1]
#     # print(cluster_rhogs_list)
#     logger.debug("There are " + str(len(cluster_rhogs_list)) + " cluster_rhogs.")
#
#     # allcc_cleaned_len = [len(i) for i in cluster_rhogs]
#     # len(allcc),len(cluster_rhogs),np.sum(allcc_cleaned_len)
#
#     return cluster_rhogs_list



def merge_rhogs2(hogmaps, rhogs_prots, conf_infer_roothogs):
    logger.debug("started merging  ")

    candidates_pair = find_rhog_candidate_pairs(hogmaps, rhogs_prots, conf_infer_roothogs) # rhogs_prots

    logger.debug(f"There are {len(candidates_pair)} candidate pairs of rhogs for merging.")
    cluster_rhogs_list = cluster_rhogs(candidates_pair)
    logger.debug(f"There are {len(cluster_rhogs_list)} clusters.")
    logger.debug("** the recursion limit is "+str( sys.getrecursionlimit()))
    cluster_rhogs_list = cluster_rhogs_nx(cluster_rhogs_list, candidates_pair)

    print("There are " + str(len(cluster_rhogs_list)) + " selected clusters.")

    # cluster_rhogs_list = cluster_rhogs(candidates_pair)

    num_rhog_g1 = 0
    for rhg, list_pr in rhogs_prots.items():
        if len(list_pr) > 1:
            num_rhog_g1 += 1

    logger.debug("There are " + str(num_rhog_g1) + " rhogs (size>1) before merging.")
    print("len(rhogs_prots) is ", len(rhogs_prots))

    file_out_merge =  open("merging_rhogs.txt","w")
    file_out_merge.write("#first column is the host hog, the rest will be merged here.\n")
    for cluster in cluster_rhogs_list:
        cluster_host = min(cluster) # write the host rhog as the first element
        file_out_merge.write(cluster_host + "\t")
        for cluster_i in cluster:
            if cluster_i != cluster_host:
                file_out_merge.write(cluster_i+"\t")
        file_out_merge.write("\n")
    file_out_merge.close()

    counter_merged_prots=0
    newhost= ""
    rhogs_host_size = 0
    for cluster in cluster_rhogs_list:
        if len(cluster)>30:
            print(len(cluster))
        #prots = [rhogs_prots[hog] for hog in cluster]
         #  cluster[0] #
        found = False

        while not found and len(cluster)>0:
            host_hog = min(cluster)
            if host_hog in rhogs_prots:
                found = True
                rhogs_host_size = len(rhogs_prots[host_hog])
            else:
                # logger print rhog
                cluster.remove(host_hog)

        all_prots = []
        all_prots2 =[]
        for hog in cluster:
            # todo check, create smallers clusters
            if len(all_prots) < conf_infer_roothogs.big_rhog_size:
                if hog in rhogs_prots:
                    all_prots += rhogs_prots[hog]
                    del rhogs_prots[hog]
            else:
                if newhost:
                    if len(all_prots2) < conf_infer_roothogs.big_rhog_size:
                        if hog in rhogs_prots:
                            all_prots2 += rhogs_prots[hog]
                            del rhogs_prots[hog]
                    else:
                        print("cluster is very big "+str(len(cluster)))
                else:
                    newhost = hog

        if newhost:
            all_prots_uniq2 = list(set(all_prots2))  # there might be repeated prots
            rhogs_prots[newhost] = all_prots_uniq2

        all_prots_uniq = list(set(all_prots))  # there might be repeated prots
        rhogs_prots[host_hog] = all_prots_uniq
        # merging D0562038 and D0559070
        # tr | C3ZG56 | C3ZG56_BRAFL HOG: D0562038, tr | H2Y1V7 | H2Y1V7_CIOIN HOG: D0562038, tr | C3ZG56 | C3ZG56_BRAFL HOG: D0559070, tr | H2Y1V7 | H2Y1V7_CIOIN HOG: D0559070
        if len(all_prots_uniq) > rhogs_host_size:
            counter_merged_prots += len(all_prots_uniq)
        # otherwise, merging didn't help

    print(len(rhogs_prots),counter_merged_prots)

    num_rhog_g1 = 0
    for rhg, list_pr in rhogs_prots.items():
        if len(list_pr) > 1:
            num_rhog_g1 += 1

    logger.debug("There are " + str(num_rhog_g1) + " rhogs (size>1)   by merging "+str(counter_merged_prots)+" proteins in total.")

    return rhogs_prots


#
# def merge_rhogs(hogmaps, rhogs_prots, conf_infer_roothogs):
#     logger.debug("started merging  ")
#
#     candidates_pair = find_rhog_candidate_pairs(hogmaps, rhogs_prots)
#
#     cluster_rhogs_list = cluster_rhogs(candidates_pair)
#     num_rhog_g1 = 0
#     for rhg, list_pr in rhogs_prots.items():
#         if len(list_pr) > 1:
#             num_rhog_g1 += 1
#
#     logger.debug("There are " + str(num_rhog_g1) + " rhogs (size>1) before merging.")
#     print(len(rhogs_prots))
#
#     file_out_merge =  open("merging_rhogs.txt","w")
#     file_out_merge.write("#first column is the host hog, the rest will be merged here.\n")
#     for cluster in cluster_rhogs_list:
#         cluster_host = min(cluster) # write the host rhog as the first element
#         file_out_merge.write(cluster_host + "\t")
#         for cluster_i in cluster:
#             if cluster_i != cluster_host:
#                 file_out_merge.write(cluster_i+"\t")
#         file_out_merge.write("\n")
#     file_out_merge.close()
#
#     counter_merged_prots=0
#     newhost= ""
#     for cluster in cluster_rhogs_list:
#         if len(cluster)>30:
#             print(len(cluster))
#         #prots = [rhogs_prots[hog] for hog in cluster]
#          #  cluster[0] #
#         found = False
#         while not found and len(cluster)>0:
#             host_hog = min(cluster)
#             if host_hog in rhogs_prots:
#                 found = True
#                 rhogs_host_size = len(rhogs_prots[host_hog])
#             else:
#                 # logger print rhog
#                 cluster.remove(host_hog)
#
#         all_prots = []
#         all_prots2 =[]
#         for hog in cluster:
#             # todo check, create smallers clusters
#             if len(all_prots) < conf_infer_roothogs.big_rhog_size:
#                 if hog in rhogs_prots:
#                     all_prots += rhogs_prots[hog]
#                     del rhogs_prots[hog]
#             else:
#                 if newhost:
#                     if len(all_prots2) < conf_infer_roothogs.big_rhog_size:
#                         if hog in rhogs_prots:
#                             all_prots2 += rhogs_prots[hog]
#                             del rhogs_prots[hog]
#                     else:
#                         print("cluster is very big "+str(len(cluster)))
#                 else:
#                     newhost = hog
#
#         if newhost:
#             all_prots_uniq2 = list(set(all_prots2))  # there might be repeated prots
#             rhogs_prots[newhost] = all_prots_uniq2
#
#         all_prots_uniq = list(set(all_prots))  # there might be repeated prots
#         rhogs_prots[host_hog] = all_prots_uniq
#         # merging D0562038 and D0559070
#         # tr | C3ZG56 | C3ZG56_BRAFL HOG: D0562038, tr | H2Y1V7 | H2Y1V7_CIOIN HOG: D0562038, tr | C3ZG56 | C3ZG56_BRAFL HOG: D0559070, tr | H2Y1V7 | H2Y1V7_CIOIN HOG: D0559070
#         if len(all_prots_uniq) > rhogs_host_size:
#             counter_merged_prots += len(all_prots_uniq)
#         # otherwise, merging didn't help
#
#     print(len(rhogs_prots),counter_merged_prots)
#
#     num_rhog_g1 = 0
#     for rhg, list_pr in rhogs_prots.items():
#         if len(list_pr) > 1:
#             num_rhog_g1 += 1
#
#     logger.debug("There are " + str(num_rhog_g1) + " rhogs (size>1)   by merging "+str(counter_merged_prots)+" proteins in total.")
#
#     return rhogs_prots


def collect_unmapped_singleton(rhogs_prots, unmapped, prot_recs_all, unmapped_singleton_fasta="singleton_unmapped.fa"):
    unmapped_recs = []
    for species_name, prot_names in unmapped.items():
        all_recs_of_species = prot_recs_all[species_name]
        for prot_name in prot_names:
            try:
                unmapped_recs.append(all_recs_of_species[prot_name])
            except KeyError:
                # some small prots are removed in the begining min_sequence_length
                pass
    logger.debug(f"collected sequence records of {len(unmapped_recs)} unmapped proteins.")

    singleton_recs = []
    for rhogid, sp_prot_list in rhogs_prots.items():
        if len(sp_prot_list) == 1:
            species_name = sp_prot_list[0][0]
            prot_name = sp_prot_list[0][1]
            prot_rec = prot_recs_all[species_name][prot_name]
            singleton_recs.append(prot_rec)
    logger.debug(f"collected sequence records of {len(singleton_recs)} singleton proteins.")

    singleton_unmapped_recs = unmapped_recs + singleton_recs
    SeqIO.write(singleton_unmapped_recs, unmapped_singleton_fasta , "fasta")
    logger.debug(f"Wrote {len(singleton_unmapped_recs)} sequences of unmapped and singleton proteins to {unmapped_singleton_fasta}.")
    return len(singleton_unmapped_recs)


def run_linclust(fasta_to_cluster="singleton_unmapped.fa"):   # todo move run_linclust to _wrapper.py .  see easy-cluster below.
    # todo: change FastOMA.nf to assgin more cpus to infer_roothog step.  mmseqs uses all cpus by default  # num_threads = 10  , --threads " + str(num_threads) + "
    command_clust = mmseqs_executable_path +" easy-cluster  " + fasta_to_cluster + " singleton_unmapped tmp_linclust"
    # easy-cluster is much better than easy-linclust but a bit slower. todo: make it as an arugment for user

    logger.debug("clustering rooting started " + command_clust)
    process = subprocess.Popen(command_clust.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output, error = process.communicate()
    logger.debug(f"clustering rooting finished with returncode {process.returncode}.")
    if process.returncode != 0:
        logger.error(f"mmseqs easy-cluster failed with returncode {process.returncode}.")
        logger.error(f"stdout:\n{output.decode('utf-8')}\n******************")
        logger.error(f"stderr:\n{error.decode('utf-8')}\n******************")
        raise RuntimeError("mmseqs easy-cluster failed.")
    return "singleton_unmapped_all_seqs.fasta"  # the fasta-like file of clusters


def write_clusters(fasta_clustered, address_rhogs_folder, min_rhog_size):
    """parse the fasta-like output of mmseqs linclust and write each cluster

    The format is described here: https://github.com/soedinglab/MMseqs2/wiki#cluster-fasta-like-format
    Example:
        >O67032                                 <- cluster 1
        >sp|O67032|RF1_AQUAE                    <- member1 of cluster 1
        MLKEAYISRLDKLQEKYRKLQEELSKPEVIQD...
        >sp|O84026|RF1_CHLTR                    <- member2 of cluster 1
        MEIKVLECLKRLEEVEKQISDPNIFSNPKEYS...
        >sp|P47500|RF1_MYCGE                    <- member3 of cluster 1
        MDFDKQLFFNVEKIVELTEQLEKDLNKPNLSF...
        >O66429                                 <- cluster 2
        >sp|O66429|EFTU_AQUAE                   <- member1 of cluster 2
        MAKEKFERTKEHVNVGTIGHVDHGKS
        >sp|P0CD71|EFTU_CHLTR                   <- member1 of cluster 1
        MSKETFQRNKPHINIGTIGHVDHGKTTLTAAITRAL
    """

    with open(fasta_clustered, 'rt') as cluster_file:
        previous_line_start = " "
        clusters = []
        cluster = []
        for line in cluster_file:
            line_strip = line.strip()
            if previous_line_start == " ":
                clusters = []
                cluster = []
            elif line_strip[0] == ">" and previous_line_start == ">":  # new cluster started at previous line
                if len(cluster) >= 5:
                    # more than one records [ID1, seq1,ID2, seq2, newcluster_id], i.e. non-trivial cluster
                    clusters.append(cluster[:-1])  # last item is the id of new cluster
                cluster = [line_strip]
            elif line_strip[0] == ">" and previous_line_start != ">":
                cluster.append(line_strip)
            elif line_strip[0] != ">":
                cluster.append(line_strip)
            previous_line_start = line_strip[0]

    # for last cluster
    if len(cluster) >= 4:  # more than one records  [ID1, seq1,ID2, seq2]
        clusters.append(cluster)
    logger.debug(f"Number of non-trivial linclust clusters raw is {len(clusters)}.")

    cluster_iter = 10000
    for cluster in clusters:
        # cluster is like ['>sp|O67032|RF1_AQUAE', 'MLKEAYISRLDKLQEKYRKLQEELSKPEVIQD...', ...], i.e. an id and its sequence.
        # thus, check for double the min_rhog_size to have at least min_rhog_size proteins
        if len(cluster) >= 2 * min_rhog_size:
            species_names = set([])     # it seems that genes from same species tend to cluster together, discard such clusters
            for pr_idi in range(int(len(cluster)/2)):
                species_name = cluster[pr_idi*2].split(" ")[0].split("||")[1]
                species_names.add(species_name)
            if len(species_names) > 1:
                with open(address_rhogs_folder + "/HOG_clust" + str(cluster_iter) + ".fa", "w") as file_idx:
                    for line in cluster:
                        file_idx.write(line + "\n")
                cluster_iter += 1
    return cluster_iter - 10000


def parse_isoform_files(species_names, folder=None):
    if folder is None:
        folder = "./splice"
    isoform_by_gene_all = {}

    for species_name in species_names:  # from fasta file
        file_splice_name = os.path.join(folder,  species_name + ".splice")
        if os.path.isfile(file_splice_name):
            # from OMArk
            isoform_by_gene = []
            with open(file_splice_name) as handle:
                for line in handle.readlines():
                    line = line.strip('\n')
                    splice = line.split(";")
                    isoform_by_gene.append(splice)
        else:
            isoform_by_gene = []
        isoform_by_gene_all[species_name] = isoform_by_gene
    return isoform_by_gene_all


def select_best_isoforms(species_names, isoform_by_gene_all, hogmaps):
    isoform_selected = {}
    isoform_not_selected = {}

    for species_name in species_names:
        isoform_not_selected[species_name] = []
        isoform_selected[species_name] = []
        isoform_list_sp = isoform_by_gene_all[species_name]
        hogmaps_sp = hogmaps[species_name]
        for isoform_list in isoform_list_sp:
            score_isof_best = 0
            protname_best = isoform_list[0]  # to keep the order if all of the isoforms' omamer score are 'na'
            for isoform_name in isoform_list:
                if isoform_name in hogmaps_sp:

                    prot_maps = hogmaps_sp[isoform_name]
                    if len(prot_maps) > 1:  # for multi-hit omamer output with -n
                        scores = [float(i[1]) for i in prot_maps]  # (hogid,score,seqlen,subfamily_medianseqlen)
                        # hogids = [i[0] for i in prot_maps]
                        # seq_lens = [i[2] for i in prot_maps]
                        # subf_med_lens = [i[3] for i in prot_maps]
                        family_score = max(scores)
                        # idx_max = scores.index(family_score)
                        # seq_len = seq_lens[idx_max]
                        # subf_med_len = subf_med_lens[idx_max]
                        # hogid = hogids[max_index]
                    else:
                        # hogid = prot_maps[0][0]
                        family_score = float(prot_maps[0][1])
                        # seq_len = prot_maps[0][2]
                        # subf_med_len = prot_maps[0][3]
                    score_isof = float(family_score) # for omamer v2 fscore is enough * min(int(seq_len), int(subf_med_len))
                    # print(isoform_name, score_isof,score_isof_best)

                    if score_isof >= score_isof_best:  # when there is a tie, the last one is selected!
                        protname_best = isoform_name
                        score_isof_best = score_isof

            isoform_selected[species_name].append(protname_best)
            isoform_not_selected[species_name].extend([i for i in isoform_list if i != protname_best])  # flatten for all

    return isoform_selected, isoform_not_selected


def write_selected_isoforms(isoform_data, prot_recs_lists, output_folder="selected_isoforms"):
    """
    write isoforms to a tsv file for each species
    """
    selected_isoforms_folder = Path(output_folder)
    selected_isoforms_folder.mkdir(exist_ok=True)

    isoform_selected = isoform_data['selected_isoforms']
    isoform_by_gene_all = isoform_data['isoform_by_gene']
    
    for species, isoform_selected_sp in isoform_selected.items():
        isoform_by_gene = isoform_by_gene_all[species]
        assert len(isoform_by_gene) == len(isoform_selected_sp)

        main_to_minor = {
            main: set(gene_isoforms) - {main}
            for main, gene_isoforms in zip(isoform_selected_sp, isoform_by_gene)
        }
        all_isoforms = set(chain.from_iterable(isoform_by_gene))
        fpath = selected_isoforms_folder / f"{species}_selected_isoforms.fa"
        with open(fpath, "wt") as fout:
            for rec in prot_recs_lists[species]:
                id_ = rec.id.split("||")[0]
                if id_ in main_to_minor or id_ not in all_isoforms:
                    if id_ in main_to_minor:
                        desc_ = "minor_isoforms=" + ",".join(main_to_minor.get(id_, []))
                    else:
                        desc_ = ""
                    fout.write(f">{id_} {desc_}\n{rec.seq}\n")


def filter_hogmaps_by_isoforms(hogmaps, non_selected_isoforms):
    hogmaps_selected_isof = {}
    for species, hogmap in hogmaps.items():
        hogmaps_selected_isof[species]={}

        non_selected_isoforms_sp = set(non_selected_isoforms[species])
        for prot in hogmap:
            if prot not in non_selected_isoforms_sp:
                hogmaps_selected_isof[species][prot]=hogmap[prot]

    return hogmaps_selected_isof


def cluster_rhogs_nx(cluster_rhogs_list, candidates_pair):
    count=0
    candidates_pair = set(candidates_pair)
    new_cluster = []
    for c in cluster_rhogs_list:
        count+=1

        G = nx.Graph()
        for i in range(len(c)):
            pairA = c[i]

            if not G.has_node(pairA):
                G.add_node(pairA)
            for j in range(i+1, len(c)):

                pairB = c[j]
                if not G.has_node(pairB):
                    G.add_node(pairB)
                if (pairA,pairB) in candidates_pair or (pairB,pairA) in candidates_pair:
                    G.add_edge(pairA,pairB)
        density = nx.density(G) # todo: do we really need this line ?
        if len(G) <1000: # todo make it as a parameter
            # limit number of merging rootHOGs due to O(N^3) of HCS/cc
            newG = HCS(G)
            clusters = [ list(x) for x in nx.connected_components(newG) if len(x)>1]
        else:
            logger.debug("This cluster of rootHOg hit the limit of 1000 rHOGs, we are not merging these rootHOGs.")
            clusters =  [ [i] for i in list(G.nodes)] # do not merge at all

        new_cluster += clusters
    return new_cluster

def highly_connected(G, E):
    """Checks if the graph G is highly connected

    Highly connected means, that splitting the graph G into subgraphs needs more than 0.5*|V| edge deletions
    This definition can be found in Section 2 of the publication.

    :param G: Graph G
    :param E: Edges needed for splitting G
    :return: True if G is highly connected, otherwise False
    """

    return len(E) >= len(G.nodes) / 2


def remove_edges(G, E):
    """Removes all edges E from G

    Iterates over all edges in E and removes them from G
    :param G: Graph to remove edges from
    :param E: One or multiple Edges
    :return: Graph with edges removed
    """

    for edge in E:
        G.remove_edge(*edge)
    return G


def HCS(G):
    """Basic HCS Algorithm
https://github.com/53RT/Highly-Connected-Subgraphs-Clustering-HCS/blob/master/hcs.py
    cluster labels, removed edges are stored in global variables

    :param G: Input graph
    :return: Either the input Graph if it is highly connected, otherwise a Graph composed of
    Subgraphs that build clusters
    """

    E = nx.algorithms.connectivity.cuts.minimum_edge_cut(G)

    if not highly_connected(G, E):
        G = remove_edges(G, E)
        sub_graphs = [G.subgraph(c).copy() for c in nx.connected_components(G)]

        if len(sub_graphs) == 2:
            H = HCS(sub_graphs[0])
            _H = HCS(sub_graphs[1])

            G = nx.compose(H, _H)

    return G



import os
import time
import tempfile
import logging
from pyparsing import ParseException
from Bio import AlignIO, SeqIO

from .base_treebuilder import TreeBuilder, AlignmentInput, DataType
from .parsers import PhymlParser

from ..abstract_cli import AbstractCLI
from ..options import StringOption, FlagOption, IntegerOption, FloatOption, MultiOption, OptionSet


logger = logging.getLogger(__name__)



class PhymlCLI(AbstractCLI):
    @property
    def _default_exe(self):
        return 'phyml'


def set_default_dna_options(treebuilder):
    """
    Dummy function as sensible default
    """
    treebuilder.options = get_default_options()
    treebuilder.options['-d'].set_value('nt')


def set_default_protein_options(treebuilder):
    """
    Dummy function as sensible default
    """
    treebuilder.options = get_default_options()


class Phyml(TreeBuilder):
    """ Phyml tree reconstruction

    This wrapper can be called to reconstruct a phylogenetic tree
    using PhyML.
    """

    def __init__(self, alignment, *args, **kwargs):
        """
        :param alignment: input multiple sequence alignment. This can be either
            a filename or an biopython SeqRecord collection.
        """
        super(Phyml, self).__init__(alignment, *args, **kwargs)
        self.options = get_default_options()
        if self.datatype == DataType.DNA:
            set_default_dna_options(self)
        else:
            set_default_protein_options(self)

    def __call__(self, *args, **kwargs):
        """
        Anything to do with calling Mafft should go here.
        If any extra arguments need to be passed they can
        be specified (listed as *args and **kwargs for now).
        """
        start = time.time()  # time the execution

        if self.input_type == AlignmentInput.OBJECT:  # different operation depending on what it is
            with tempfile.NamedTemporaryFile(mode='wt') as fh:
                SeqIO.write(self.input, fh, 'phylip-relaxed')  # default interleaved
                fh.seek(0)
                output, error = self._call(fh.name, *args, **kwargs)
                self.result = self._read_result(fh.name)  # store result
        else:
            path = os.path.dirname(self.input)
            filename = os.path.basename(self.input)
            # some operations done because phyml can not deal with large filenames that are caused due to a large path
            with os.chdir(path):
                output, error = self._call(filename, *args, **kwargs)
                self.result = self._read_result(filename)  # store result

        self.stdout = output
        self.stderr = error

        end = time.time()
        self.elapsed_time = end - start
        return self.result["tree"]
        # End call

    # Any other accessory methods
    def _call(self, filename, *args, **kwargs):
        """
        Call underlying low level _Phyml wrapper.
        Options are passed via *args and **kwargs
        [This only covers the simplest automatic
         case]
        """
        self.cli('{} -i {}'.format(self.command(), filename),
                 wait=True)
        return self.cli.get_stdout(), self.cli.get_stderr()

    def command(self):
        return str(self.options)

    def _read_result(self, output):
        """
        Read back the result.
        """

        #TODO: change the output dictionary into a better format
        expected_outfiles = ['{}_phyml_stats'.format(output), '{}_phyml_tree'.format(output)]
        parser = PhymlParser()

        # Phyml outputs two outfiles, a stats file and a tree file.
        # Sometimes it appends .txt, sometimes not. Seems to be platform-specific.
        # Here we assume they are without .txt, but if we can't find them, try
        # looking for the .txt onees instead
        try:
            # Check if these are the .txt style outfiles
            if not os.path.exists(expected_outfiles[0]):
                expected_outfiles = [x + '.txt' for x in expected_outfiles]
            result = parser.to_dict(*expected_outfiles)

        except IOError as ioerr:
            logger.error('Error reading results')
            result = None
        except ParseException as parseerr:
            logger.error('Other parse error', parseerr)
            result = None

        return result

    def _init_cli(self, binary):
        return PhymlCLI(executable=binary)


def get_default_options():
    return OptionSet([
        # Algorithm

        # Set datatype to nt or aa
        StringOption('-d', 'aa', active=True),

        # Set the model for either DNA or AA alignment
        StringOption('-m', '', active=False),

        # If set to true will assume sequential format
        FlagOption('-q', False, active=False),

        # Set bootstrap value
        IntegerOption('-b', 0, active=False),

        # Tree topology search operation option
        StringOption('-s', 'NNI', active=False)
    ])



import os
import time
import tempfile
import logging


from pyparsing import ParseException
from Bio import AlignIO, SeqIO

from .parsers import ProtTestParser
from .base_modeltester import ModelTester, AlignmentInput, DataType

from ..abstract_cli import AbstractCLI
from ..options import StringOption, FlagOption, OptionSet

logger = logging.getLogger(__name__)


class ProtTestCLI(AbstractCLI):
    """
    Especially in this case it is important that the $PROTTEST_HOME environmental variable is set to the installation directory of the prottest tool
    """
    @property
    def _default_exe(self):
        return 'java -jar ' + os.environ['PROTTEST_HOME'] + '/prottest-3.4.2.jar'


def set_default_dna_options(modeltester):
    """
    Dummy function as sensible default
    """
    modeltester.options = get_default_options()


def set_default_protein_options(modeltester):
    """
    Dummy function as sensible default
    """
    modeltester.options = get_default_options()


class ProtTest(ModelTester):
    """ ProtTest to determine the best model for a specific alignment
    This wrapper can be called to test various models for phylogeny inference.
    """

    def __init__(self, alignment, *args, **kwargs):
        """
        :param alignment: input multiple sequence alignment. This can be either
            a filename or an biopython SeqRecord collection.
        """
        self.options = get_default_options()
        super(ProtTest, self).__init__(alignment=alignment, *args, **kwargs)
        if self.datatype == DataType.DNA:
            set_default_dna_options(self)
        else:
            set_default_protein_options(self)

    def __call__(self, *args, **kwargs):
        """
        Anything to do with calling ProtTest should go here.
        If any extra arguments need to be passed they can
        be specified (listed as *args and **kwargs for now).
        """
        start = time.time()  # time the execution
        if self.input_type == AlignmentInput.OBJECT:  # different operation depending on what it is
            with tempfile.NamedTemporaryFile(mode='wt') as filehandle:
                SeqIO.write(self.input, filehandle, 'fasta')
                filehandle.seek(0)
                output, error = self._call(filehandle.name, *args, **kwargs)
        else:
            output, error = self._call(self.input, *args, **kwargs)

        self.result = self._read_result(output)  # store result
        self.stdout = output
        self.stderr = error

        end = time.time()
        self.elapsed_time = end - start
        return self.result
        # End call

    # Any other accessory methods
    def _call(self, filename, *args, **kwargs):
        """
        Call underlying low level _ProtTest wrapper.
        Options are passed via *args and **kwargs
        [This only covers the simplest automatic
         case]
        """
        self.cli('{} -i {}'.format(self.command(), filename),
                 wait=True)
        return self.cli.get_stdout(), self.cli.get_stderr()

    def command(self):
        return str(self.options)

    def _read_result(self, output):

        parser = ProtTestParser()

        try:
            result = parser.to_dict(output)

        except IOError as ioerr:
            logger.error('Error reading results')
            result = None
        except ParseException as parseerr:
            logger.error('Other parse error', parseerr)
            result = None

        return result


    def _init_cli(self, binary):
        return ProtTestCLI(executable=binary)


def get_default_options():
    return OptionSet([
        # Algorithm

        # Display models sorted by Akaike Information Criterion (AIC)
        FlagOption('-AIC', False, active=False),

        # Display models sorted by Decision Theory Criterion
        FlagOption('-DT', False, active=False),

        # Tree file (optional) [default: NJ tree]
        StringOption('-t', '', active=False),

        # Display models sorted by Corrected Akaike Information Criterion (AICc)
        FlagOption('-AICC', False, active=False),

        #Enables / Disables PhyML logging into log directory(see prottest.properties)
        FlagOption('-log', False, active=False)
    ])




import logging
import os
import time

from Bio import SeqIO, AlignIO
from pyparsing import ParseException
import tempfile

from .base_trimmer import MSATrimmer
from ..aligners.base_aligner import AlignmentInput
from ...wrappers import WrapperError
from ..abstract_cli import AbstractCLI
from ..options import OptionSet, StringOption, IntegerOption, FlagOption, FloatOption
from ...seq_utils.utils import iter_seqrecs_from_any
from ...file_utils import TempFile, TempDir

logger = logging.getLogger(__name__)


class TrimAlCLI(AbstractCLI):
    """
       TrimAl low-level command line interface

       :Example:

       ::

           trimal_cli = TrimAlCLI()
           process = trimal_cli(cmd='trimal args...')
           stdout = trimal_cli.get_stdout()
       """
    @property
    def _default_exe(self):
        return 'trimal'

    @property
    def _hyphen_policy(self):
        return 1


def set_representative_options(trimmer, size:int):
    """
    get most representative sequences for msa of size `size`
    """
    trimmer.options = get_default_options()
    trimmer.options['-cluster'].set_value(size)


def get_phylogenetic_ml_options(trimmer):
    """Use an heuristic to decide the optimal method for trimming the alignment.
    (see User Guide for details).
    Optimized for Maximum Likelihood phylogenetic tree reconstruction"""
    trimmer.options = get_default_options()
    trimmer.options['-automated1'].set_value(True)


def get_phylogenetic_neighbour_joining(trimmer):
    """Use automated selection on "strictplus" mode. (see User Guide).
    Optimized for Neighbour Joining phylogenetic tree reconstruction. """
    trimmer.options = get_default_options()
    trimmer.options['-strictplus'].set_value(True)


class TrimAl(MSATrimmer):
    """
    Convenient wrapper for TrimAl MAS Trimmer

    The wrapper is written as a callable class.
    This can hold data (state) to do with the operation it performs, so it can keep results,
    execution times and other metadata, as well as perform the task.

    This is a basic implementation that can be extended. The important parts are
    __init__ (does the setup) and __call__ (does the work). All
    else are helper methods.

    :Example:

    ::

        callable_wrapper = TrimAl(aln)
        result = callable_wrapper()
        time_taken = callable_wrapper.elapsed_time
        result_again = callable_wrapper.result


    .. note:: There exists an ipython notebook on how to work with wrappers,
         including dealing with non-default parameters.
    """

    def __init__(self, input_, *args, **kwargs):
        super(TrimAl, self).__init__(input_, *args, **kwargs)
        self.options = get_default_options()

    def __call__(self, *args, **kwargs):
        """
        Anything to do with calling Mafft should go here.
        If any extra arguments need to be passed they can
        be specified (listed as *args and **kwargs for now).
        """
        start = time.time()  # time the execution

        with tempfile.TemporaryDirectory() as tmpdir:
            outfn = os.path.join(tmpdir, "output.trimal.fasta")
            if self.input_type == AlignmentInput.OBJECT:  # different operation depending on what it is
                with open(os.path.join(tmpdir, "sequences.fa"), 'wt') as fh:
                    SeqIO.write(iter_seqrecs_from_any(self.input), fh, 'fasta')
                output, error = self._call(fh.name, outfn, *args, **kwargs)
            else:
                output, error = self._call(self.input, outfn, *args, **kwargs)
            self.result = self._read_result(outfn)
        logger.debug('Output of TrimAl: stdout={}; stderr={}'.format(output, error))
        if len(output) == 0 and len(error) > 0:
            raise WrapperError('TrimAl did not compute any alignments. StdErr: {}'.format(error))
        self.stdout = output
        self.stderr = error

        end = time.time()
        self.elapsed_time = end - start
        return self.result
        # End call

    # Any other accessory methods
    def _call(self, filename, outfn, *args, **kwargs):
        """
        Call underlying low level _Mafft wrapper.
        Options are passed via *args and **kwargs
        [This only covers the simplest automatic
         case]
        """
        if self.command() == "-fasta":
            logger.warning("no trimming options specified. will use '-automated1' insted.")
            self.options['-automated1'].set_value(True)

        self.cli('-in {} -out {} {}'.format(filename, outfn, self.command()),
                 wait=True)
        return self.cli.get_stdout(), self.cli.get_stderr()

    def command(self):
        return str(self.options)

    def _read_result(self, output):
        """
        Read back the result.
        """
        return AlignIO.read(output, 'fasta')

    def _init_cli(self, binary):
        return TrimAlCLI(executable=binary)


def get_default_options():
    return OptionSet([
        # Algorithm

        # Use a heuristic selection of the automatic method based on
        # similarity statistics. (see User Guide).
        # (Optimized for Maximum Likelihood phylogenetic tree reconstruction).
        FlagOption('-automated1', False, active=False),

        # Only columns out of internal boundaries (first and last column without gaps) are
        # candidated to be trimmed depending on the applied method
        FlagOption('-terminalonly', False, active=False),

        # Use automated selection on "strictplus" mode. (see User Guide).
        # Optimized for Neighbour Joining phylogenetic tree reconstruction.
        FlagOption('-strictplus', False, active=False),

        # Use automated selection on "gappyout" mode. This method only uses information
        # based on gaps' distribution. (see User Guide).
        FlagOption('-gappyout', False, active=False),

        # Use automated selection on "strict" mode. (see User Guide).
        FlagOption('-strict', False, active=False),

        # Get the relationship between the columns in the old and new alignment.
        FlagOption('-colnumbering', False, active=False),

        # Get the complementary alignment
        FlagOption('-complementary', False, active=False),

        # Minimum column block size to be kept in the trimmed alignment. Available with
        # manual and automatic (gappyout) methods
        IntegerOption('-block', 30, active=False),

        # -gt -gapthreshold <n>    1 - (fraction of sequences with a gap allowed). Range: [0 - 1]
        FloatOption('-gapthreshold', 0.5, active=False),

        # Minimum average similarity allowed. Range: [0 - 1]
        FloatOption('-simthreshold', 0.5, active=False),

        # Minimum percentage of the positions in the original alignment to conserve. Range: [0 - 100]
        IntegerOption('-cons', 50, active=False),

        # Format of output file should be fasta
        FlagOption('-fasta', True, active=True),

        # Use a user-defined scoring matrix.  Default: BLOSUM62
        StringOption('-matrix', '', active=False),
    ])






import logging


def setup_logging(verbosity: int, thread_names: bool = False):
    """
    Configure logging based on verbosity levels:
    -v   : INFO level on main package, WARNING on sub-packages
    -vv  : DEBUG level on main package, INFO on sub-packages
    -vvv : DEBUG on all packages
    """
    log_levels = {
        0: (logging.WARNING, logging.WARNING),  # Default: WARNING everywhere
        1: (logging.INFO, logging.WARNING),     # -v
        2: (logging.DEBUG, logging.INFO),       # -vv
        3: (logging.DEBUG, logging.DEBUG),      # -vvv
    }

    # Get the appropriate levels
    main_level, sub_level = log_levels.get(max(0, verbosity), (logging.DEBUG, logging.DEBUG))

    # Set main package logger
    logging.getLogger("FastOMA").setLevel(main_level)

    # Set sub-packages
    logging.getLogger("FastOMA.zoo").setLevel(sub_level)
    if thread_names:
        fmt = "%(asctime)s %(levelname)-8s [%(threadName)s] %(name)s: %(message)s"
    else:
        fmt = "%(asctime)s %(levelname)-8s %(name)s: %(message)s"

    # Configure root logger
    logging.basicConfig(
        format=fmt,
        level=min(main_level, sub_level),  # Ensure root captures the lowest level needed
    )




import os
import random
import types
from collections import defaultdict
from enum import Enum
from pathlib import Path


from Bio import AlignIO
from Bio.Align import MultipleSeqAlignment
from Bio.Data.IUPACData import ambiguous_dna_letters
from Bio.Seq import Seq # , UnknownSeq
from Bio.SeqRecord import SeqRecord

__all__ = ['bootstrap', 'concatenate', 'is_dna', 'BackTranslator', 'identify_input', 'AlignmentInput',
           'iter_seqrecs_from_any']


def is_dna(obj):
    """check whether a sequence is of type dna"""
    if isinstance(obj, SeqRecord):
        return is_dna(str(obj.seq))
    elif isinstance(obj, Seq):
        return is_dna(str(obj))
    elif isinstance(obj, str):
        chars = [char.upper() for char in obj if char.upper() not in '-?X']
        return set(chars).issubset(set(ambiguous_dna_letters))
    else:
        return all(is_dna(z) for z in obj)


def iter_seqrecs_from_any(obj):
    if isinstance(obj, SeqRecord):
        yield obj
    elif isinstance(obj, str):
        yield SeqRecord(Seq(obj))
    elif isinstance(obj, Seq):
        yield SeqRecord(obj)
    else:
        for sub in obj:
            yield from iter_seqrecs_from_any(sub)

def _sample_wr(population, k):
    _int = int
    _random = random.random
    n = len(population)
    return [population[_int(_random() * n)] for _ in range(k)]


def bootstrap(alignment):
    length = alignment.get_alignment_length()
    columns = [alignment[:, n] for n in range(length)]
    sample = sorted(_sample_wr(columns, length))
    sequences = [''.join(x) for x in zip(*sample)]
    ids = [seqrec.id for seqrec in alignment]
    return _assemble_msa(sequences, ids)


def _assemble_msa(strings, ids):
    seqs = []
    for (s, i) in zip(strings, ids):
        seq = Seq(s)
        seqs.append(SeqRecord(seq, id=i))
    return MultipleSeqAlignment(seqs)


def concatenate(alignments):
    """
    Concatenates a list of multiple sequence alignment objects.

    The alignments are concatenated based on their label, i.e. the
    sequences from the different alignments which have the same id/labels
    will become a single sequence. The order is preserved.

    If any sequences are missing in one or several alignments, these parts
    are padded with unknown data (:py:class:`Bio.Seq.UnknownSeq`).

    :param alignments: the list of alignments objects, i.e. list(:py:class:`Bio.Align.MultipleSeqAlignment`)
    :returns: a single :py:class:`Bio.Align.MultipleSeqAlignment`

    Example::

        >>> sequences = {'aln1': {'seq1': 'acgtca',
        ...                       'seq2': 'acgtt-',
        ...                       'seq3': 'ac-ta-'},
        ...              'aln2': {'seq2': 'ttg-cta',
        ...                       'seq3': 'tcgacta',
        ...                       'seq4': 'ttgacta'}}
        >>> alignments = [MultipleSeqAlignment([SeqRecord(Seq(sequence), id=key, annotations={"molecule_type": "DNA"})
        ...      for (key, sequence) in sequences[aln].items()])
        ...               for aln in ('aln1', 'aln2')]
        >>> con_alignment = concatenate(alignments)
        >>> con_alignment.sort()
        >>> print(con_alignment)
        ExtendedIUPACDNA() alignment with 4 rows and 13 columns
        acgtcaNNNNNNN seq1
        acgtt-ttg-cta seq2
        ac-ta-tcgacta seq3
        NNNNNNttgacta seq4

    :note:

       Limitations: any annotations in the sub-alignments are lost in
       the concatenated alignment.

    """

    # First check to see whether we're inputting filenames of alignments or the Biopython alignments
    # Assume that it's a biopython alignment if it's not a filename
    tmp_aligns = []
    for filename in alignments:
        if identify_input(filename).name == 'FILENAME':
            tmp_aligns.append(AlignIO.read(filename, "fasta"))
        else:
            tmp_aligns.append(filename)

    # Copy back to alignments
    alignments = tmp_aligns

    # Get the full set of labels (i.e. sequence ids) for all the alignments
    all_labels = set(seq.id for aln in alignments for seq in aln)

    # Make a dictionary to store info as we go along
    # (defaultdict is convenient -- asking for a missing key gives back an empty list)
    tmp = defaultdict(list)

    # try to get molecule_type from sequences
    molecule_type = set(seq.annotations.get('molecule_type') for aln in alignments for seq in aln)
    molecule_type.discard(None)
    if len(molecule_type) == 1:
        molecule_type = molecule_type.pop()
        if molecule_type.upper() in ("DNA", "RNA"):
            unknown_char = "N"
        elif molecule_type.lower() == "protein":
            unknown_char = "X"
        else:
            unknown_char = "?"
    else:
        unknown_char = '?'

    for aln in alignments:
        length = aln.get_alignment_length()

        # check if any labels are missing in the current alignment
        these_labels = set(rec.id for rec in aln)
        missing = all_labels - these_labels

        # if any are missing, create unknown data of the right length,
        # stuff the string representation into the tmp dict
        for label in missing:
            new_seq = unknown_char*length  # UnknownSeq(length, character=unknown_char) # deprecate fro biopython
            tmp[label].append(str(new_seq))

        # else stuff the string representation into the tmp dict
        for rec in aln:
            tmp[rec.id].append(str(rec.seq))

    # Stitch all the substrings together using join (most efficient way),
    # and build the Biopython data structures Seq, SeqRecord and MultipleSeqAlignment
    return MultipleSeqAlignment(SeqRecord(Seq(''.join(v)), id=k)
                                for (k, v) in tmp.items())


AlignmentInput = Enum('AlignmentInput', 'OBJECT FILENAME')


def identify_input(alignment):
    """
    Work out if we're dealing with a Biopython object (return True), a file
    (return False), or invalid input (raise error)
    """
    try:
        if isinstance(alignment, (MultipleSeqAlignment, types.GeneratorType, list)):
            # `alignment` is a Biopython MultipleSequenceAlignment
            return AlignmentInput.OBJECT
        elif hasattr(alignment, '__next__') and hasattr(alignment, '__iter__'):
            # alignmet is an iterator
            return AlignmentInput.OBJECT
        elif isinstance(alignment, (str, Path)) and os.path.exists(alignment):
            # `alignment` is a filepath
            return AlignmentInput.FILENAME
    except:
        # `alignment` is some other thing we can't handle
        raise ValueError('{} is not an alignment object or a valid filename'.format(alignment))


class BackTranslator(object):
    """
    Reverse translate a protein sequence into a DNA sequence using a
    codon usage table
    """

    def __init__(self, table):
        """
        Initialises the amino acid -> codons, codon weights map from
        the supplied codon usage table -- see http://www.kazusa.or.jp/codon
        """
        self.map = self._codon_table_parser(table)

    def _weighted_choice(self, choices):
        total = sum(w for c, w in choices)
        r = random.uniform(0, total)
        upto = 0
        for c, w in choices:
            if upto + w > r:
                return c
            upto += w
        assert False, "Shouldn't get here"

    def _read_entry(self, gen):
        """
        Reads a single entry from the active codon usage table
        """
        codon = next(gen)
        aa = next(gen)
        next(gen)
        next(gen)
        count = next(gen)
        if count == '(':
            count = next(gen)
        count = int(count.strip('()'))
        return codon.replace('U', 'T'), aa, count

    def _codon_table_parser(self, table):
        """
        Parses the table produced by http://www.kazusa.or.jp/codon
        Returns a dict of amino acid -> codons, codon weights
        """
        table = iter(table.replace('\n', '  ').split())
        codons = defaultdict(list)
        counts = defaultdict(list)
        try:
            while table:
                codon, aa, count = self._read_entry(table)
                codons[aa].append(codon)
                counts[aa].append(count)
        except StopIteration:
            pass

        wmap = {}
        for k in counts:
            s = sum(counts[k])
            scaled = [c / float(s) for c in counts[k]]
            wmap[k] = (codons[k], scaled)
        return wmap

    def back_translate(self, seq):
        """
        Back translate a protein sequence into dna. Codons are chosen randomly according
        to the codon usage frequencies in the table the BackTranslator object was 
        initialised with -- see http://www.kazusa.or.jp/codon
        """
        return ''.join(self._weighted_choice(list(zip(*self.map[char.upper()]))) for char in seq)
